# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import division

import logging
import os
from typing import Set, List

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_pages import WikipediaPages
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType, \
    WikipediaSearchMethod
from nlp_architect.data.cdc_resources.wikipedia.wiki_elastic import WikiElastic
from nlp_architect.data.cdc_resources.wikipedia.wiki_offline import WikiOffline
from nlp_architect.data.cdc_resources.wikipedia.wiki_online import WikiOnline
from nlp_architect.utils.string_utils import StringUtils

logger = logging.getLogger(__name__)


class WikipediaRelationExtraction(RelationExtraction):
    def __init__(self, method: WikipediaSearchMethod = WikipediaSearchMethod.ONLINE,
                 wiki_file: str = None, host: str = None, port: int = None,
                 index: str = None, filter_pronouns: bool = True,
                 filter_time_data: bool = True) -> None:
        """
        Extract Relation between two mentions according to Wikipedia knowledge

        Args:
            method (optional): WikipediaSearchMethod.{ONLINE/OFFLINE/ELASTIC} run against wiki
                site a sub-set of wiki or on a local elastic database (default = ONLINE)
            wiki_file (required on OFFLINE mode): str Location of Wikipedia file to work with
            host (required on Elastic mode): str the Elastic search host name
            port (required on Elastic mode): int the Elastic search port number
            index (required on Elastic mode): int the Elastic search index name
        """
        logger.info('Loading Wikipedia module')
        self.filter_pronouns = filter_pronouns
        self.filter_time_data = filter_time_data
        connectivity = method
        if connectivity == WikipediaSearchMethod.ONLINE:
            self.pywiki_impl = WikiOnline()
        elif connectivity == WikipediaSearchMethod.OFFLINE:
            if wiki_file is not None and os.path.isdir(wiki_file):
                self.pywiki_impl = WikiOffline(wiki_file)
            else:
                raise FileNotFoundError('Wikipedia resource file not found or not in path, '
                                        'create it or change the initialization method')
        elif connectivity == WikipediaSearchMethod.ELASTIC:
            self.pywiki_impl = WikiElastic(host, port, index)

        logger.info('Wikipedia module lead successfully')
        super(WikipediaRelationExtraction, self).__init__()

    def get_phrase_related_pages(self, mention_str: str) -> WikipediaPages:
        """
        Get all WikipediaPages pages related with this mention string

        Args:
            mention_str: str

        Returns:
            WikipediaPages
        """
        pages = self.pywiki_impl.get_pages(mention_str.strip())
        ret_pages = WikipediaPages()
        if pages:
            for search_page in pages:
                if search_page.page_result.pageid != 0:
                    ret_pages.add_page(search_page.page_result)

        return ret_pages

    def extract_all_relations(self, mention_x: MentionDataLight,
                              mention_y: MentionDataLight) -> Set[RelationType]:
        """
        Try to find if mentions has anyone or more of the relations this class support

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight

        Returns:
            Set[RelationType]: One or more of: RelationType.WIKIPEDIA_BE_COMP,
                RelationType.WIKIPEDIA_TITLE_PARENTHESIS,
                RelationType.WIKIPEDIA_DISAMBIGUATION, RelationType.WIKIPEDIA_CATEGORY,
                RelationType.WIKIPEDIA_REDIRECT_LINK, RelationType.WIKIPEDIA_ALIASES,
                RelationType.WIKIPEDIA_PART_OF_SAME_NAME
        """
        relations = set()
        mention1_str = mention_x.tokens_str.strip()
        mention2_str = mention_y.tokens_str.strip()

        if self.filter_pronouns:
            if self.is_both_opposite_personal_pronouns(mention1_str, mention2_str):
                relations.add(RelationType.NO_RELATION_FOUND)
                return relations

        if self.filter_time_data:
            if self.is_both_data_or_time(mention_x, mention_y):
                relations.add(RelationType.NO_RELATION_FOUND)
                return relations

        pages1 = self.get_phrase_related_pages(mention1_str)
        pages2 = self.get_phrase_related_pages(mention2_str)

        # check if search phrase is empty meaning it is probably a stop word
        if pages1.is_empty_norm_phrase or pages2.is_empty_norm_phrase:
            relations.add(RelationType.NO_RELATION_FOUND)
            return relations

        if self.is_redirect_same(pages1, pages2):
            relations.add(RelationType.WIKIPEDIA_REDIRECT_LINK)

        titles1 = pages1.get_and_set_titles()
        titles1.add(mention1_str + ' ' + mention2_str)
        titles1.add(mention2_str + ' ' + mention1_str)

        titles2 = pages2.get_and_set_titles()
        titles2.add(mention1_str + ' ' + mention2_str)
        titles2.add(mention2_str + ' ' + mention1_str)

        relation_alias = self.extract_aliases(pages1, pages2, titles1, titles2)
        if relation_alias is not RelationType.NO_RELATION_FOUND:
            relations.add(relation_alias)
        relation_dis = self.extract_disambig(pages1, pages2, titles1, titles2)
        if relation_dis is not RelationType.NO_RELATION_FOUND:
            relations.add(relation_dis)
        relation_cat = self.extract_category(pages1, pages2, titles1, titles2)
        if relation_cat is not RelationType.NO_RELATION_FOUND:
            relations.add(relation_cat)
        relation_par = self.extract_parenthesis(pages1, pages2, titles1, titles2)
        if relation_par is not RelationType.NO_RELATION_FOUND:
            relations.add(relation_par)
        relation_be = self.extract_be_comp(pages1, pages2, titles1, titles2)
        if relation_be is not RelationType.NO_RELATION_FOUND:
            relations.add(relation_be)

        if len(relations) == 0:
            relations.add(RelationType.NO_RELATION_FOUND)

        return relations

    def extract_sub_relations(self, mention_x: MentionDataLight, mention_y: MentionDataLight,
                              relation: RelationType) -> RelationType:
        """
        Check if input mentions has the given relation between them

        Args:
            mention_x: MentionDataLight
            mention_y: MentionDataLight
            relation: RelationType

        Returns:
            RelationType: relation in case mentions has given relation or
                RelationType.NO_RELATION_FOUND otherwise
        """
        mention1_str = mention_x.tokens_str.strip()
        mention2_str = mention_y.tokens_str.strip()

        if self.filter_pronouns:
            if self.is_both_opposite_personal_pronouns(mention1_str, mention2_str):
                return RelationType.NO_RELATION_FOUND

        if self.filter_time_data:
            if self.is_both_data_or_time(mention_x, mention_y):
                return RelationType.NO_RELATION_FOUND

        pages1 = self.get_phrase_related_pages(mention1_str)
        pages2 = self.get_phrase_related_pages(mention2_str)

        # check if search phrase is empty meaning it is probably a stop word
        if pages1.is_empty_norm_phrase or pages2.is_empty_norm_phrase:
            return RelationType.NO_RELATION_FOUND

        if relation == RelationType.WIKIPEDIA_REDIRECT_LINK:
            if self.is_redirect_same(pages1, pages2):
                return RelationType.WIKIPEDIA_REDIRECT_LINK

            return RelationType.NO_RELATION_FOUND

        titles1 = pages1.get_and_set_titles()
        titles1.add(mention1_str + ' ' + mention2_str)
        titles1.add(mention2_str + ' ' + mention1_str)

        titles2 = pages2.get_and_set_titles()
        titles2.add(mention1_str + ' ' + mention2_str)
        titles2.add(mention2_str + ' ' + mention1_str)

        if relation == RelationType.WIKIPEDIA_ALIASES:
            return self.extract_aliases(pages1, pages2, titles1, titles2)
        if relation == RelationType.WIKIPEDIA_DISAMBIGUATION:
            return self.extract_disambig(pages1, pages2, titles1, titles2)
        if relation == RelationType.WIKIPEDIA_CATEGORY:
            return self.extract_category(pages1, pages2, titles1, titles2)
        if relation == RelationType.WIKIPEDIA_TITLE_PARENTHESIS:
            return self.extract_parenthesis(pages1, pages2, titles1, titles2)
        if relation == RelationType.WIKIPEDIA_BE_COMP:
            return self.extract_be_comp(pages1, pages2, titles1, titles2)

        return RelationType.NO_RELATION_FOUND

    @staticmethod
    def extract_be_comp(pages1: WikipediaPages, pages2: WikipediaPages, titles1: Set[str],
                        titles2: Set[str]) -> RelationType:
        """
        Check if input mentions has be-comp/is-a relation

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage
            titles1: Set[str]
            titles2: Set[str]

        Returns:
            RelationType.WIKIPEDIA_BE_COMP or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        if bool(pages1.get_and_set_be_comp() & titles2):
            relation = RelationType.WIKIPEDIA_BE_COMP
        elif bool(pages2.get_and_set_be_comp() & titles1):
            relation = RelationType.WIKIPEDIA_BE_COMP
        return relation

    @staticmethod
    def extract_parenthesis(pages1: WikipediaPages, pages2: WikipediaPages, titles1: Set[str],
                            titles2: Set[str]) -> RelationType:
        """
        Check if input mentions has parenthesis relation

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage
            titles1: Set[str]
            titles2: Set[str]

        Returns:
            RelationType.WIKIPEDIA_TITLE_PARENTHESIS or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        if bool(pages1.get_and_set_parenthesis() & titles2):
            relation = RelationType.WIKIPEDIA_TITLE_PARENTHESIS
        elif bool(pages2.get_and_set_parenthesis() & titles1):
            relation = RelationType.WIKIPEDIA_TITLE_PARENTHESIS
        return relation

    @staticmethod
    def extract_category(pages1: WikipediaPages, pages2: WikipediaPages, titles1: Set[str],
                         titles2: Set[str]) -> RelationType:
        """
        Check if input mentions has category relation

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage
            titles1: Set[str]
            titles2: Set[str]

        Returns:
            RelationType.WIKIPEDIA_CATEGORY or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        if bool(pages1.get_and_set_all_categories() & titles2):
            relation = RelationType.WIKIPEDIA_CATEGORY
        elif bool(pages2.get_and_set_all_categories() & titles1):
            relation = RelationType.WIKIPEDIA_CATEGORY
        return relation

    @staticmethod
    def extract_disambig(pages1: WikipediaPages, pages2: WikipediaPages, titles1: Set[str],
                         titles2: Set[str]) -> RelationType:
        """
        Check if input mentions has disambiguation relation

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage
            titles1: Set[str]
            titles2: Set[str]

        Returns:
            RelationType.WIKIPEDIA_DISAMBIGUATION or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        if bool(pages1.get_and_set_all_disambiguation() & titles2):
            relation = RelationType.WIKIPEDIA_DISAMBIGUATION
        elif bool(pages2.get_and_set_all_disambiguation() & titles1):
            relation = RelationType.WIKIPEDIA_DISAMBIGUATION
        return relation

    @staticmethod
    def extract_aliases(pages1: WikipediaPages, pages2: WikipediaPages, titles1: Set[str],
                        titles2: Set[str]) -> RelationType:
        """
        Check if input mentions has aliases relation

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage
            titles1: Set[str]
            titles2: Set[str]
        Returns:
            RelationType.WIKIPEDIA_ALIASES or RelationType.NO_RELATION_FOUND
        """
        relation = RelationType.NO_RELATION_FOUND
        if bool(pages1.get_and_set_all_aliases() & titles2):
            relation = RelationType.WIKIPEDIA_ALIASES
        elif bool(pages2.get_and_set_all_aliases() & titles1):
            relation = RelationType.WIKIPEDIA_ALIASES
        return relation

    def is_part_of_same_name(self, pages1: WikipediaPages, pages2: WikipediaPages) -> bool:
        """
        Check if input mentions has part of same name relation (eg: page1=John, page2=Smith)

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage

        Returns:
            bool
        """
        for page1 in pages1.pages:
            for page2 in pages2.pages:
                if page1.relations.is_part_name and page2.relations.is_part_name:
                    pages = self.pywiki_impl.get_pages(page1.orig_phrase + ' ' + page2.orig_phrase)
                    for page in pages:
                        if page.page_result.pageid != 0:
                            return True
        return False

    @staticmethod
    def is_redirect_same(pages1: WikipediaPages, pages2: WikipediaPages) -> bool:
        """
        Check if input mentions has same wikipedia redirect page

        Args:
            pages1: WikipediaPages
            pages2: WikipediaPage

        Returns:
            bool
        """
        for page1 in pages1.get_pages():
            for page2 in pages2.get_pages():
                if page1.pageid > 0 and page2.pageid > 0:
                    if page1.pageid == page2.pageid:
                        return True
        return False

    @staticmethod
    def get_supported_relations() -> List[RelationType]:
        """
        Return all supported relations by this class

        Returns:
            List[RelationType]
        """
        return [RelationType.WIKIPEDIA_BE_COMP, RelationType.WIKIPEDIA_TITLE_PARENTHESIS,
                RelationType.WIKIPEDIA_DISAMBIGUATION, RelationType.WIKIPEDIA_CATEGORY,
                RelationType.WIKIPEDIA_REDIRECT_LINK, RelationType.WIKIPEDIA_ALIASES,
                RelationType.WIKIPEDIA_PART_OF_SAME_NAME]

    @staticmethod
    def is_both_opposite_personal_pronouns(phrase1: str, phrase2: str) -> bool:
        """
        check if both phrases refers to pronouns

        Returns:
            bool
        """
        result = False
        if StringUtils.is_pronoun(phrase1.lower()) and StringUtils.is_pronoun(phrase2.lower()):
            result = True

        return result

    @staticmethod
    def is_both_data_or_time(mention1: MentionDataLight, mention2: MentionDataLight) -> bool:
        """
        check if both phrases refers to time or date

        Returns:
            bool
        """
        mention1_ner = mention1.mention_ner
        mention2_ner = mention2.mention_ner

        if mention1_ner is None:
            _, _, _, mention1_ner = StringUtils.find_head_lemma_pos_ner(mention1.tokens_str)
        if mention2_ner is None:
            _, _, _, mention2_ner = StringUtils.find_head_lemma_pos_ner(mention2.tokens_str)

        is1_time_or_data = 'DATE' in mention1_ner or 'TIME' in mention1_ner
        is2_time_or_data = 'DATE' in mention2_ner or 'TIME' in mention2_ner

        result = False
        if is1_time_or_data and is2_time_or_data:
            result = True

        return result



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import time

from nlp_architect.common.cdc.cluster import Clusters
from nlp_architect.common.cdc.topics import Topic
from nlp_architect.models.cross_doc_coref.system.sieves.sieves import SieveClusterMerger
from nlp_architect.models.cross_doc_coref.system.sieves_container_init import \
    SievesContainerInitialization

logger = logging.getLogger(__name__)


class RunSystemsSuper(object):
    def __init__(self, topic: Topic):
        self.sieves = []
        self.results_dict = dict()
        self.results_ordered = []
        logger.info('loading topic %s, total mentions: %d', topic.topic_id, len(topic.mentions))
        self.clusters = Clusters(topic.topic_id, topic.mentions)

    @staticmethod
    def set_sieves_from_config(config, get_rel_extraction):
        sieves = []
        for _type_tup in config.sieves_order:
            sieves.append(SieveClusterMerger(_type_tup, get_rel_extraction(_type_tup[0])))
        return sieves

    def run_deterministic(self):
        # pylint: disable=too-many-nested-blocks
        for sieve in self.sieves:
            start = time.time()
            clusters_changed = True
            merge_count = 0
            while clusters_changed:
                clusters_changed = False
                clusters_size = len(self.clusters.clusters_list)
                for i in range(0, clusters_size):
                    cluster_i = self.clusters.clusters_list[i]
                    if cluster_i.merged:
                        continue

                    for j in range(i + 1, clusters_size):
                        cluster_j = self.clusters.clusters_list[j]
                        if cluster_j.merged:
                            continue

                        if cluster_i is not cluster_j:
                            criterion = sieve.run_sieve(cluster_i, cluster_j)
                            if criterion:
                                merge_count += 1
                                clusters_changed = True
                                cluster_i.merge_clusters(cluster_j)
                                cluster_j.merged = True

                if clusters_changed:
                    self.clusters.clean_clusters()

            end = time.time()
            took = end - start
            logger.info('Total of %d clusters merged using method: %s, took: %.4f sec',
                        merge_count, str(sieve.excepted_relation), took)

        return self.clusters

    def get_results(self):
        return self.results_ordered


class RunSystemsEntity(RunSystemsSuper):
    def __init__(self, topic: Topic, resources):
        super(RunSystemsEntity, self).__init__(topic)
        self.sieves = self.set_sieves_from_config(resources.entity_config,
                                                  resources.get_module_from_relation)


class RunSystemsEvent(RunSystemsSuper):
    def __init__(self, topic, resources):
        super(RunSystemsEvent, self).__init__(topic)
        self.sieves = self.set_sieves_from_config(resources.event_config,
                                                  resources.get_module_from_relation)


# pylint: disable=no-else-return
def get_run_system(topic: Topic, resource: SievesContainerInitialization, eval_type: str):
    if eval_type.lower() == 'entity':
        return RunSystemsEntity(topic, resource)
    if eval_type.lower() == 'event':
        return RunSystemsEvent(topic, resource)
    else:
        raise AttributeError(eval_type + ' Not supported!')



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import argparse
import csv
import os
import logging
import sys
from multiprocessing import Pool
from os import path, makedirs

from fastText import train_unsupervised
from newspaper import Article

from nlp_architect.solutions.trend_analysis.np_scorer import NPScorer
from nlp_architect import LIBRARY_OUT
from nlp_architect.utils.io import validate_existing_directory
from nlp_architect.utils.text import SpacyInstance

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)
data_dir = str(LIBRARY_OUT / 'trend-analysis-data')


def noun_phrase_extraction(docs, parser):
    """
    Extract noun-phrases from a textual corpus

    Args:
        docs (List[String]): A list of documents
        parser (SpacyInstance): Spacy NLP parser
    Returns:
        List of topics with their tf_idf, c_value, language-model scores
    """
    logger.info('extracting NPs')
    np_app = NPScorer(parser=parser)
    np_result = np_app.score_documents(docs, return_all=True)
    logger.info('got NP extractor result. %s phrases were extracted', str(len(np_result)))
    return np_result


def load_text_from_folder(folder):
    """
    Load files content into a list of docs (texts)

    Args:
        folder: A path to a folder containing text files

    Returns:
        A list of documents (List[String])
    """
    files_content = []
    try:
        for filename in os.listdir(folder):
            with open(folder + os.sep + filename, "rb") as f:
                try:
                    files_content.append(
                        ((f.read()).decode('UTF-8', errors='replace')).replace(
                            '\n', ' ').replace('\r',
                                               ' ').replace(
                            '\b', ' '))
                except Exception as e:
                    logger.error(str(e))
    except Exception as e:
        logger.error("Error in load_text: %s", str(e))

    return files_content


def load_url_content(url_list):
    """
    Load articles content into a list of docs (texts)

    Args:
        url_list (List[String]): A list of urls

    Returns:
        A list of documents (List[String])
    """
    files_content = []
    url = ''
    try:
        for url in url_list:
            try:
                url = str(url)
                logger.info("loading %s", url)
                article = Article(url)
                article.download()
                article.parse()
                files_content.append(article.title + ' ' + article.text)
            except Exception as e:
                logger.error(str(e))
    except Exception as e:
        logger.error("Error in load_text: %s, for url: %s", str(e), str(url))

    return files_content


def save_scores(np_result, file_path):
    """
    Save the result of a topic extraction into a file

    Args:
        np_result: A list of topics with different score types (tfidf, cvalue, freq)
        file_path: The output file path
    """
    logger.info('saving multi-scores np extraction results to: %s', file_path)
    with open(file_path, 'wt', encoding='utf-8', newline='') as csv_file:
        writer = csv.writer(csv_file, delimiter=',')
        ctr = 0
        for group, tfidf, cvalue, freq in np_result:
            try:
                group_str = ""
                for p in group:
                    group_str += p + ';'
                row = (group_str[0:-1], str(tfidf), str(cvalue), str(freq))
                writer.writerow(row)
                ctr += 1
            except Exception as e:
                logger.error("Error while writing np results. iteration #: %s. Error: %s", str(
                    ctr), str(e))


def unify_corpora_from_texts(text_list_t, text_list_r):
    """
    Merge two corpora into a single text file

    Args:
        text_list_t: A list of documents - target corpus (List[String])
        text_list_r: A list of documents - reference corpus (List[String])
    Returns:
        The path of the unified corpus
    """
    logger.info('prepare data for w2v training')
    out_file = str(path.join(data_dir, 'corpus.txt'))
    try:
        with open(out_file, 'w+', encoding='utf-8') as writer:
            write_text_list_to_file(text_list_t, writer)
            write_text_list_to_file(text_list_r, writer)
    except Exception as e:
        logger.error('Error: %s', str(e))
    return out_file


def unify_corpora_from_folders(corpus_a, corpus_b):
    """
    Merge two corpora into a single text file

    Args:
        corpus_a: A folder containing text files (String)
        corpus_b: A folder containing text files (String)
    Returns:
        The path of the unified corpus
    """
    logger.info('prepare data for w2v training')
    out_file = str(path.join(data_dir, 'corpus.txt'))
    try:
        with open(out_file, 'w+', encoding='utf-8') as writer:
            write_folder_corpus_to_file(corpus_a, writer)
            write_folder_corpus_to_file(corpus_b, writer)
    except Exception as e:
        logger.error('Error: %s', str(e))
    return out_file


def write_folder_corpus_to_file(corpus, writer):
    """
    Merge content of a folder into a single text file

    Args:
        corpus: A folder containing text files (String)
        writer: A file writer
    """
    for filename in os.listdir(corpus):
        try:
            file_path = str(path.join(corpus, filename))
            with open(file_path, "r", encoding='utf-8') as f:
                lines = f.readlines()
                writer.writelines(lines)
        except Exception as e:
            logger.error('Error: %s. skipping file: %s', str(e), str(filename))


def write_text_list_to_file(text_list, writer):
    """
    Merge a list of texts into a single text file

    Args:
        text_list: A list of texts (List[String])
        writer: A file writer
    """
    try:
        for text in text_list:
            writer.writelines(text)
    except Exception as e:
        logger.error('Error: %s', str(e))


def train_w2v_model(data):
    """
    Train a w2v (skipgram) model using fasttext package

    Args:
        data: A path to the training data (String)
    """
    logger.info('Fasttext embeddings training...')
    try:
        model = train_unsupervised(input=data, model='skipgram', epoch=100, minCount=1, dim=100)
        model.save_model(str(path.join(data_dir, 'W2V_Models/model.bin')))
    except Exception as e:
        logger.error('Error: %s', str(e))


def create_w2v_model(text_list_t, text_list_r):
    """
     Create a w2v model on the given corpora

     Args:
         text_list_t: A list of documents - target corpus (List[String])
         text_list_r: A list of documents - reference corpus (List[String])
     """
    unified_data = unify_corpora_from_texts(text_list_t, text_list_r)
    train_w2v_model(unified_data)


def get_urls_from_file(file):
    """
     Merge two corpora into a single text file

     Args:
         corpus_a: A folder containing text files (String)
         corpus_b: A folder containing text files (String)
     Returns:
         The path of the unified corpus
     """
    with open(file) as url_file:
        url_list = url_file.readlines()
        url_list = [x.strip() for x in url_list]
    return url_list


def initiate_parser():
    return SpacyInstance(
        disable=['tagger', 'ner', 'parser', 'vectors', 'textcat']).parser


def main(corpus_t, corpus_r, single_thread, no_train, url):
    try:
        if not path.exists(data_dir):
            makedirs(data_dir)
        if not path.exists(path.join(data_dir, 'W2V_Models')):
            os.makedirs(path.join(data_dir, 'W2V_Models'))
    except Exception:
        logger.error('failed to create output folder.')
        sys.exit()
    if url:
        if not path.isfile(corpus_t) or not path.isfile(corpus_r):
            logger.error('Please provide a valid csv file with urls')
            sys.exit()
        else:
            text_t = load_url_content(get_urls_from_file(corpus_t))
            text_r = load_url_content(get_urls_from_file(corpus_r))
    else:
        if not path.isdir(corpus_t) or not path.isdir(corpus_r):
            logger.error('Please provide valid directories for target_corpus and'
                         ' for ref_corpus')
            sys.exit()

        else:
            text_t = load_text_from_folder(corpus_t)
            text_r = load_text_from_folder(corpus_r)

    # extract noun phrases
    nlp_parser_t = initiate_parser()
    nlp_parser_r = initiate_parser()
    if single_thread:
        result_t = noun_phrase_extraction(text_t, nlp_parser_t)
        result_r = noun_phrase_extraction(text_r, nlp_parser_r)
    else:
        with Pool(processes=2) as pool:
            run_np_t = pool.apply_async(noun_phrase_extraction,
                                        [text_t, nlp_parser_t])
            run_np_r = pool.apply_async(noun_phrase_extraction,
                                        [text_r, nlp_parser_r])
            result_t = run_np_t.get()
            result_r = run_np_r.get()

    # save results to csv

    # save_scores(result_t, os.path.splitext(os.path.basename(corpus_t))[0] + '.csv')
    # save_scores(result_r, os.path.splitext(os.path.basename(corpus_r))[0] + '.csv')
    save_scores(result_t, str(path.join(data_dir, os.path.basename(corpus_t))) + '.csv')
    save_scores(result_r, str(path.join(data_dir, os.path.basename(corpus_r))) + '.csv')

    # create w2v model
    if not no_train:
        create_w2v_model(text_t, text_r)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(prog='topic_extraction.py')
    argparser.add_argument('target_corpus', metavar='target_corpus',
                           type=validate_existing_directory,
                           help='a path to a folder containing text files')
    argparser.add_argument('ref_corpus', metavar='ref_corpus', type=validate_existing_directory,
                           help='a path to a folder containing text files')
    argparser.add_argument('--no_train', action='store_true',
                           help='skip the creation of w2v model')
    argparser.add_argument('--url', action='store_true',
                           help='corpus provided as csv file with urls')
    argparser.add_argument('--single_thread', action='store_true',
                           help='analyse corpora sequentially')
    args = argparser.parse_args()

    main(args.target_corpus, args.ref_corpus, args.single_thread, args.no_train, args.url)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
from os import listdir
from os.path import join, isfile

from nlp_architect.data.cdc_resources.data_types.wn.wordnet_page import WordnetPage
from nlp_architect.utils.io import load_json_file

logger = logging.getLogger(__name__)


class WordnetOffline(object):
    def __init__(self, wordnet_dump):
        if wordnet_dump:
            self.dump = self.load_dump(wordnet_dump)
            logger.info('Wikipedia dump loaded successfully!')

    def get_pages(self, mention):
        page = None
        if mention.tokens_str is not None and mention.tokens_str in self.dump:
            page = self.dump[mention.tokens_str]

        return page

    def load_dump(self, wn_dump):
        onlyfiles = []
        for _file in listdir(wn_dump):
            file_path = join(wn_dump, _file)
            if isfile(file_path):
                onlyfiles.append(file_path)

        json_dump_list = {}
        for _file in onlyfiles:
            json_dump_list.update(load_json_file(_file))

        dump_final = {}
        for key, value in json_dump_list.items():
            dump_final[key] = self.extract_json_values(value)

        return dump_final

    @staticmethod
    def extract_json_values(json_page):
        if json_page is not None:
            orig_phrase = json_page.get('orig_phrase', None)
            clean_phrase = json_page.get('clean_phrase', None)
            head = json_page.get('head', None)
            head_lemma = json_page.get('head_lemma', None)
            head_synonyms = set(json_page.get('head_synonyms', None))
            head_lemma_synonyms = set(json_page.get('head_lemma_synonyms', None))
            head_derivationally = set(json_page.get('head_derivationally', None))
            head_lemma_derivationally = set(json_page.get('head_lemma_derivationally', None))

            all_clean_words_synonyms = json_page.get('all_clean_words_synonyms', None)
            all_as_set_list = []
            for list_ in all_clean_words_synonyms:
                all_as_set_list.append(set(list_))

            wordnet_page = WordnetPage(orig_phrase, clean_phrase, head,
                                       head_lemma, head_synonyms,
                                       head_lemma_synonyms, head_derivationally,
                                       head_lemma_derivationally,
                                       all_as_set_list)
            return wordnet_page

        return None



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import pickle
from os import path

import numpy as np
import spacy
from spacy.tokens import Doc
from spacy.tokens import Span

from nlp_architect.models.chunker import SequenceChunker
from nlp_architect.utils.generic import pad_sentences
from nlp_architect.utils.io import validate_existing_filepath
from nlp_architect.utils.text import extract_nps, Stopwords


class NPAnnotator(object):
    """
    Spacy based NP annotator - uses models.SequenceChunker model for annotation

    Args:
        model (SequenceChunker): a chunker model
        word_vocab (Vocabulary): word-id vocabulary of the model
        char_vocab (Vocabulary): char id vocabulary of words of the model
        chunk_vocab (Vocabulary): chunk tag vocabulary of the model
        batch_size (int, optional): inference batch size
    """

    def __init__(self, model, word_vocab, char_vocab, chunk_vocab, batch_size: int = 32):
        self.model = model
        self.bs = batch_size
        self.word_vocab = word_vocab
        self.char_vocab = char_vocab
        self.chunk_vocab = chunk_vocab
        Doc.set_extension('noun_phrases', default=[], force=True)

    @classmethod
    def load(cls, model_path: str, parameter_path: str, batch_size: int = 32,
             use_cudnn: bool = False):
        """
        Load a NPAnnotator annotator

        Args:
            model_path (str): path to trained model
            parameter_path (str): path to model parameters
            batch_size (int, optional): inference batch_size
            use_cudnn (bool, optional): use gpu for inference (cudnn cells)

        Returns:
            NPAnnotator class with loaded model
        """

        _model_path = path.join(path.dirname(path.realpath(__file__)), model_path)
        validate_existing_filepath(_model_path)
        _parameter_path = path.join(path.dirname(path.realpath(__file__)), parameter_path)
        validate_existing_filepath(_parameter_path)

        model = SequenceChunker(use_cudnn=use_cudnn)
        model.load(_model_path)
        with open(_parameter_path, 'rb') as fp:
            model_params = pickle.load(fp)
            word_vocab = model_params['word_vocab']
            chunk_vocab = model_params['chunk_vocab']
            char_vocab = model_params.get('char_vocab', None)
        return cls(model, word_vocab, char_vocab, chunk_vocab, batch_size)

    def _infer_chunks(self, input_vec, doc_lengths):
        tagged_sents = self.model.predict(input_vec, batch_size=self.bs).argmax(2)
        sentence = []
        for c, l in zip(tagged_sents, doc_lengths):
            sentence.append(c[-l:])
        doc = np.concatenate(sentence)
        chunk_tags = [self.chunk_vocab.id_to_word(w) for w in doc]
        return extract_nps(chunk_tags)

    def _feature_extractor(self, doc):
        features = np.asarray([self.word_vocab[w] if self.word_vocab[w] is not None else 1
                               for w in doc])
        if self.char_vocab:
            sentence_chars = []
            for w in doc:
                word_chars = []
                for c in w:
                    _cid = self.char_vocab[c]
                    word_chars.append(_cid if _cid is not None else 1)
                sentence_chars.append(word_chars)
            sentence_chars = pad_sentences(sentence_chars, self.model.max_word_len)
            features = (features, sentence_chars)
        return features

    def __call__(self, doc: Doc) -> Doc:
        """
        Annotate the document with noun phrase spans
        """
        spans = []
        doc_vecs = []
        doc_chars = []
        doc_lens = []
        if len(doc) < 1:
            return doc
        for sentence in doc.sents:
            features = self._feature_extractor([t.text for t in sentence])
            if isinstance(features, tuple):
                doc_vec = features[0]
                doc_chars.append(features[1])
            else:
                doc_vec = features
            doc_vecs.append(doc_vec)
            doc_lens.append(len(doc_vec))
        doc_vectors = pad_sentences(np.asarray(doc_vecs))
        inputs = doc_vectors
        if self.char_vocab:
            max_len = doc_vectors.shape[1]
            padded_chars = np.zeros((len(doc_chars), max_len, self.model.max_word_len))
            for idx, d in enumerate(doc_chars):
                d = d[:max_len]
                padded_chars[idx, -d.shape[0]:] = d
            inputs = [inputs, padded_chars]
        np_indexes = self._infer_chunks(inputs, doc_lens)
        for s, e in np_indexes:
            np_span = Span(doc, s, e)
            spans.append(np_span)
        spans = _NPPostprocessor.process(spans)
        set_noun_phrases(doc, spans)
        return doc


def get_noun_phrases(doc: Doc) -> [Span]:
    """
    Get noun phrase tags from a spacy annotated document.

    Args:
        doc (Doc): a spacy type document

    Returns:
        a list of noun phrase Span objects
    """
    assert hasattr(doc._, 'noun_phrases'), 'no noun_phrase attributes in document'
    return doc._.noun_phrases


def set_noun_phrases(doc: Doc, nps: [Span]) -> None:
    """
    Set noun phrase tags

    Args:
        doc (Doc): a spacy type document
        nps ([Span]): a list of Spans
    """
    assert hasattr(doc._, 'noun_phrases'), 'no noun_phrase attributes in document'
    doc._.set('noun_phrases', nps)


class _NPPostprocessor:
    @classmethod
    def process(cls, noun_phrases: [Span]) -> [Span]:
        new_phrases = []
        for phrase in noun_phrases:
            p = _NPPostprocessor._phrase_process(phrase)
            if p is not None and len(p) > 0:
                new_phrases.append(p)
        return new_phrases

    @classmethod
    def _phrase_process(cls, phrase: Span) -> Span:
        last_phrase = None
        while phrase != last_phrase:
            last_phrase = phrase
            for func_args in post_processing_rules:
                pf = func_args[0]
                args = func_args[1:]
                if len(args) > 0:
                    phrase = pf(phrase, *args)
                else:
                    phrase = pf(phrase)
                if phrase is None:
                    break
        return phrase


def _filter_repeating_nonalnum(phrase, length):
    """
    Check if a given phrase has non repeating alphanumeric chars
    of given length.
    Example: 'phrase $$$' with length=3 will return False
    """
    if len(phrase) > 0:
        alnum_len = length
        for t in phrase:
            if not t.is_alpha:
                alnum_len -= 1
            else:
                alnum_len = length
            if alnum_len == 0:
                return None
    return phrase


def _filter_long_phrases(phrase, word_length, phrase_length):
    if len(phrase) > 0 and max([len(t) for t in phrase]) > word_length \
            and len(phrase) > phrase_length:
        return None
    return phrase


def _remove_non_alphanum_from_start(phrase):
    if len(phrase) > 1 and not phrase[0].is_alpha:
        phrase = phrase[1:]
    return phrase


def _remove_non_alphanum_from_end(phrase):
    if len(phrase) > 1 and not phrase[-1].is_alpha:
        phrase = phrase[:-1]
    return phrase


def _remove_stop_words(phrase):
    while len(phrase) > 0 and (phrase[0].is_stop
                               or str(phrase[0]).strip().lower() in Stopwords.get_words()):
        phrase = phrase[1:]
    while len(phrase) > 0 and (phrase[-1].is_stop
                               or str(phrase[-1]).strip().lower() in Stopwords.get_words()):
        phrase = phrase[:-1]
    return phrase


def _remove_char_at_start(phrase):
    chars = ['@', '-', '=', '.', ':', '+', '?', 'nt', '\"', '\'', '\'S', '\'s', ',']
    if phrase and len(phrase) > 0:
        while len(phrase) > 0 and phrase[0].text in chars:
            phrase = phrase[1:]
    return phrase


def _remove_char_at_end(phrase):
    chars = [',', '(', ')', ' ', '-']
    if phrase:
        while len(phrase) > 0 and phrase[-1].text in chars:
            phrase = phrase[:-1]
    return phrase


def _remove_pos_from_start(phrase):
    tag_list = ['WDT', 'PRP$', ':']
    pos_list = ['PUNCT', 'INTJ', 'NUM', 'PART', 'ADV', 'DET', 'PRON', 'VERB']
    if phrase:
        while len(phrase) > 0 and (phrase[0].pos_ in pos_list or phrase[0].tag_ in tag_list):
            phrase = phrase[1:]
    return phrase


def _remove_pos_from_end(phrase):
    tag_list = ['WDT', ':']
    pos_list = ['DET', 'PUNCT', 'CONJ']
    if phrase:
        while len(phrase) > 0 and (phrase[-1].pos_ in pos_list or phrase[-1].tag_ in tag_list):
            phrase = phrase[:-1]
    return phrase


def _filter_single_pos(phrase):
    pos_list = ['VERB', 'ADJ', 'ADV']
    if phrase and len(phrase) == 1 and phrase[0].pos_ in pos_list:
        return None
    return phrase


def _filter_fp_nums(phrase):
    if len(phrase) > 0:
        try:
            # check for float number
            float(phrase.text.replace(',', ''))
            return None
        except ValueError:
            return phrase
    return phrase


def _filter_single_char(phrase):
    if phrase and len(phrase) == 1 and len(phrase[0]) == 1:
        return None
    return phrase


def _filter_empty(phrase):
    if phrase is None or len(phrase) == 0 or len(phrase.text) == 0 \
            or len(str(phrase.text).strip()) == 0:
        return None
    return phrase


post_processing_rules = [
    (_filter_single_char,),
    (_filter_single_pos,),
    (_remove_pos_from_start,),
    (_remove_pos_from_end,),
    (_remove_stop_words,),
    (_remove_non_alphanum_from_start,),
    (_remove_non_alphanum_from_end,),
    (_filter_repeating_nonalnum, 5),
    (_filter_long_phrases, 5, 75),
    (_remove_char_at_start,),
    (_remove_char_at_end,),
    (_filter_fp_nums,),
    (_filter_empty,),
]


class SpacyNPAnnotator(object):
    """
    Simple Spacy pipe with NP extraction annotations
    """

    def __init__(self, model_path, settings_path, spacy_model='en', batch_size=32,
                 use_cudnn=False):
        _model_path = path.join(path.dirname(path.realpath(__file__)), model_path)
        validate_existing_filepath(_model_path)
        _settings_path = path.join(path.dirname(path.realpath(__file__)), settings_path)
        validate_existing_filepath(_settings_path)

        nlp = spacy.load(spacy_model)
        for p in nlp.pipe_names:
            if p not in ['tagger']:
                nlp.remove_pipe(p)
        nlp.add_pipe(nlp.create_pipe('sentencizer'), first=True)
        nlp.add_pipe(NPAnnotator.load(_model_path, settings_path, batch_size=batch_size,
                                      use_cudnn=use_cudnn), last=True)
        self.nlp = nlp

    def __call__(self, text: str) -> [str]:
        """
        Parse a given text and return a list of noun phrases found

        Args:
            text (str): a text string

        Returns:
            list of noun phrases as strings
        """
        return [np.text for np in get_noun_phrases(self.nlp(text))]



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=no-name-in-module
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments
# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches

import random
import time
from collections import namedtuple
from operator import itemgetter

from dynet import ParameterCollection, AdamTrainer, LSTMBuilder, tanh, logistic, rectify, cmult, \
    SimpleRNNBuilder, concatenate, np, renew_cg, esum

from nlp_architect.data.conll import ConllEntry
from nlp_architect.models.bist import decoder
from nlp_architect.models.bist.utils import read_conll


# Things that were changed from the original:
# - Added input validation
# - Updated function and object names to dyNet 2.0.2 and Python 3
# - Removed external embeddings option
# - Reformatted code and variable names to conform with PEP8
# - Added dict_to_obj()
# - Added option for train() to get ConllEntry input
# - Added legal header
# - Disabled some style checks


class MSTParserLSTM(object):
    """Underlying LSTM model for MSTParser used by BIST parser."""

    def __init__(self, vocab, w2i, pos, rels, options):
        if isinstance(options, dict):
            options = _dict_to_obj(options, 'Values')

        self.model = ParameterCollection()
        random.seed(1)
        self.trainer = AdamTrainer(self.model)

        self.activations = {'tanh': tanh, 'sigmoid': logistic, 'relu': rectify,
                            'tanh3': (lambda x: tanh(cmult(cmult(x, x), x)))}
        self.activation = self.activations[options.activation]

        self.blstm_flag = options.blstmFlag
        self.labels_flag = options.labelsFlag
        self.costaug_flag = options.costaugFlag
        self.bibi_flag = options.bibiFlag

        self.ldims = options.lstm_dims
        self.wdims = options.wembedding_dims
        self.pdims = options.pembedding_dims
        self.rdims = options.rembedding_dims
        self.layers = options.lstm_layers
        self.words_count = vocab
        self.vocab = {word: ind + 3 for word, ind in list(w2i.items())}
        self.pos = {word: ind + 3 for ind, word in enumerate(pos)}
        self.rels = {word: ind for ind, word in enumerate(rels)}
        self.irels = rels

        if self.bibi_flag:
            self.builders = [LSTMBuilder(1, self.wdims + self.pdims, self.ldims, self.model),
                             LSTMBuilder(1, self.wdims + self.pdims, self.ldims, self.model)]
            self.bbuilders = [LSTMBuilder(1, self.ldims * 2, self.ldims, self.model),
                              LSTMBuilder(1, self.ldims * 2, self.ldims, self.model)]
        elif self.layers > 0:
            self.builders = \
                [LSTMBuilder(self.layers, self.wdims + self.pdims, self.ldims, self.model),
                 LSTMBuilder(self.layers, self.wdims + self.pdims, self.ldims, self.model)]
        else:
            self.builders = [SimpleRNNBuilder(1, self.wdims + self.pdims, self.ldims, self.model),
                             SimpleRNNBuilder(1, self.wdims + self.pdims, self.ldims, self.model)]

        self.hidden_units = options.hidden_units
        self.hidden2_units = options.hidden2_units

        self.vocab['*PAD*'] = 1
        self.pos['*PAD*'] = 1

        self.vocab['*INITIAL*'] = 2
        self.pos['*INITIAL*'] = 2

        self.wlookup = self.model.add_lookup_parameters((len(vocab) + 3, self.wdims))
        self.plookup = self.model.add_lookup_parameters((len(pos) + 3, self.pdims))
        self.rlookup = self.model.add_lookup_parameters((len(rels), self.rdims))

        self.hid_layer_foh = self.model.add_parameters((self.hidden_units, self.ldims * 2))
        self.hid_layer_fom = self.model.add_parameters((self.hidden_units, self.ldims * 2))
        self.hid_bias = self.model.add_parameters((self.hidden_units))

        self.hid2_layer = self.model.add_parameters((self.hidden2_units, self.hidden_units))
        self.hid2_bias = self.model.add_parameters((self.hidden2_units))

        self.out_layer = self.model.add_parameters(
            (1, self.hidden2_units if self.hidden2_units > 0 else self.hidden_units))

        if self.labels_flag:
            self.rhid_layer_foh = self.model.add_parameters((self.hidden_units, 2 * self.ldims))
            self.rhid_layer_fom = self.model.add_parameters((self.hidden_units, 2 * self.ldims))
            self.rhid_bias = self.model.add_parameters((self.hidden_units))
            self.rhid2_layer = self.model.add_parameters((self.hidden2_units, self.hidden_units))
            self.rhid2_bias = self.model.add_parameters((self.hidden2_units))
            self.rout_layer = self.model.add_parameters(
                (len(self.irels),
                 self.hidden2_units if self.hidden2_units > 0 else self.hidden_units))
            self.rout_bias = self.model.add_parameters((len(self.irels)))

    def _get_expr(self, sentence, i, j):
        # pylint: disable=missing-docstring
        if sentence[i].headfov is None:
            sentence[i].headfov = self.hid_layer_foh.expr() * concatenate(
                [sentence[i].lstms[0], sentence[i].lstms[1]])
        if sentence[j].modfov is None:
            sentence[j].modfov = self.hid_layer_fom.expr() * concatenate(
                [sentence[j].lstms[0], sentence[j].lstms[1]])

        if self.hidden2_units > 0:
            output = \
                self.out_layer.expr() * self.activation(
                    self.hid2_bias.expr() + self.hid2_layer.expr() * self.activation(
                        sentence[i].headfov + sentence[j].modfov
                        + self.hid_bias.expr()))  # + self.outBias
        else:
            output = self.out_layer.expr() * self.activation(
                sentence[i].headfov + sentence[j].modfov + self.hid_bias.expr())  # + self.outBias
        return output

    def _evaluate(self, sentence):
        # pylint: disable=missing-docstring
        exprs = [[self._get_expr(sentence, i, j) for j in range(len(sentence))]
                 for i in range(len(sentence))]
        scores = np.array([[output.scalar_value() for output in exprsRow] for exprsRow in exprs])
        return scores, exprs

    def _evaluate_label(self, sentence, i, j):
        # pylint: disable=missing-docstring
        if sentence[i].rheadfov is None:
            sentence[i].rheadfov = self.rhid_layer_foh.expr() * concatenate(
                [sentence[i].lstms[0], sentence[i].lstms[1]])
        if sentence[j].rmodfov is None:
            sentence[j].rmodfov = self.rhid_layer_fom.expr() * concatenate(
                [sentence[j].lstms[0], sentence[j].lstms[1]])

        if self.hidden2_units > 0:
            output = self.rout_layer.expr() * self.activation(
                self.rhid2_bias.expr() + self.rhid2_layer.expr()
                * self.activation(sentence[i].rheadfov + sentence[j].rmodfov
                                  + self.rhid_bias.expr())) + self.rout_bias.expr()
        else:
            output = self.rout_layer.expr() * self.activation(
                sentence[i].rheadfov + sentence[j].rmodfov
                + self.rhid_bias.expr()) + self.rout_bias.expr()
        return output.value(), output

    def predict(self, conll_path=None, conll=None):
        # pylint: disable=missing-docstring
        if conll is None:
            conll = read_conll(conll_path)

        for sentence in conll:
            conll_sentence = [entry for entry in sentence if
                              isinstance(entry, ConllEntry)]

            for entry in conll_sentence:
                wordvec = self.wlookup[int(
                    self.vocab.get(entry.norm, 0))] if self.wdims > 0 else None
                posvec = self.plookup[
                    int(self.pos[entry.pos])] if self.pdims > 0 else None
                entry.vec = concatenate(
                    [_f for _f in [wordvec, posvec, None] if _f])

                entry.lstms = [entry.vec, entry.vec]
                entry.headfov = None
                entry.modfov = None

                entry.rheadfov = None
                entry.rmodfov = None

            if self.blstm_flag:
                lstm_forward = self.builders[0].initial_state()
                lstm_backward = self.builders[1].initial_state()

                for entry, rentry in zip(conll_sentence, reversed(conll_sentence)):
                    lstm_forward = lstm_forward.add_input(entry.vec)
                    lstm_backward = lstm_backward.add_input(rentry.vec)

                    entry.lstms[1] = lstm_forward.output()
                    rentry.lstms[0] = lstm_backward.output()

                if self.bibi_flag:
                    for entry in conll_sentence:
                        entry.vec = concatenate(entry.lstms)

                    blstm_forward = self.bbuilders[0].initial_state()
                    blstm_backward = self.bbuilders[1].initial_state()

                    for entry, rentry in zip(conll_sentence, reversed(conll_sentence)):
                        blstm_forward = blstm_forward.add_input(entry.vec)
                        blstm_backward = blstm_backward.add_input(rentry.vec)

                        entry.lstms[1] = blstm_forward.output()
                        rentry.lstms[0] = blstm_backward.output()

            scores, _ = self._evaluate(conll_sentence)
            heads = decoder.parse_proj(scores)

            for entry, head in zip(conll_sentence, heads):
                entry.pred_parent_id = head
                entry.pred_relation = '_'

            dump = False

            if self.labels_flag:
                for modifier, head in enumerate(heads[1:]):
                    scores, _ = self._evaluate_label(conll_sentence, head, modifier + 1)
                    conll_sentence[modifier + 1].pred_relation = \
                        self.irels[max(enumerate(scores), key=itemgetter(1))[0]]

            renew_cg()
            if not dump:
                yield sentence

    def train(self, conll_path):
        # pylint: disable=invalid-name
        # pylint: disable=missing-docstring
        eloss = 0.0
        mloss = 0.0
        eerrors = 0
        etotal = 0
        start = time.time()

        shuffled_data = list(read_conll(conll_path))
        random.shuffle(shuffled_data)
        errs = []
        lerrs = []
        i_sentence = 0

        for sentence in shuffled_data:
            if i_sentence % 100 == 0 and i_sentence != 0:
                print('Processing sentence number:', i_sentence, 'Loss:',
                      eloss / etotal, 'Errors:',
                      (float(eerrors)) / etotal, 'Time', time.time() - start)
                start = time.time()
                eerrors = 0
                eloss = 0.0
                etotal = 0

            conll_sentence = [entry for entry in sentence if isinstance(entry, ConllEntry)]

            for entry in conll_sentence:
                c = float(self.words_count.get(entry.norm, 0))
                drop_flag = (random.random() < (c / (0.25 + c)))
                wordvec = self.wlookup[int(self.vocab.get(entry.norm, 0)) if drop_flag else 0] \
                    if self.wdims > 0 else None
                posvec = self.plookup[int(self.pos[entry.pos])] if self.pdims > 0 else None

                entry.vec = concatenate([_f for _f in [wordvec, posvec, None] if _f])

                entry.lstms = [entry.vec, entry.vec]
                entry.headfov = None
                entry.modfov = None

                entry.rheadfov = None
                entry.rmodfov = None

            if self.blstm_flag:
                lstm_forward = self.builders[0].initial_state()
                lstm_backward = self.builders[1].initial_state()

                for entry, rentry in zip(conll_sentence, reversed(conll_sentence)):
                    lstm_forward = lstm_forward.add_input(entry.vec)
                    lstm_backward = lstm_backward.add_input(rentry.vec)

                    entry.lstms[1] = lstm_forward.output()
                    rentry.lstms[0] = lstm_backward.output()

                if self.bibi_flag:
                    for entry in conll_sentence:
                        entry.vec = concatenate(entry.lstms)

                    blstm_forward = self.bbuilders[0].initial_state()
                    blstm_backward = self.bbuilders[1].initial_state()

                    for entry, rentry in zip(conll_sentence, reversed(conll_sentence)):
                        blstm_forward = blstm_forward.add_input(entry.vec)
                        blstm_backward = blstm_backward.add_input(rentry.vec)

                        entry.lstms[1] = blstm_forward.output()
                        rentry.lstms[0] = blstm_backward.output()

            scores, exprs = self._evaluate(conll_sentence)
            gold = [entry.parent_id for entry in conll_sentence]
            heads = decoder.parse_proj(scores, gold if self.costaug_flag else None)

            if self.labels_flag:
                for modifier, head in enumerate(gold[1:]):
                    rscores, rexprs = self._evaluate_label(conll_sentence, head, modifier + 1)
                    gold_label_ind = self.rels[conll_sentence[modifier + 1].relation]
                    wrong_label_ind = max(((label, scr) for label, scr in enumerate(rscores)
                                           if label != gold_label_ind), key=itemgetter(1))[0]
                    if rscores[gold_label_ind] < rscores[wrong_label_ind] + 1:
                        lerrs.append(rexprs[wrong_label_ind] - rexprs[gold_label_ind])

            e = sum([1 for h, g in zip(heads[1:], gold[1:]) if h != g])
            eerrors += e
            if e > 0:
                loss = [(exprs[h][i] - exprs[g][i]) for i, (h, g) in
                        enumerate(zip(heads, gold)) if h != g]  # * (1.0/float(e))
                eloss += e
                mloss += e
                errs.extend(loss)

            etotal += len(conll_sentence)

            if i_sentence % 1 == 0 or errs > 0 or lerrs:
                if errs or lerrs:
                    eerrs = (esum(errs + lerrs))  # * (1.0/(float(len(errs))))
                    eerrs.scalar_value()
                    eerrs.backward()
                    self.trainer.update()
                    errs = []
                    lerrs = []

                renew_cg()

            i_sentence += 1

        if errs:
            eerrs = (esum(errs + lerrs))  # * (1.0/(float(len(errs))))
            eerrs.scalar_value()
            eerrs.backward()
            self.trainer.update()

            renew_cg()

        self.trainer.update()
        print("Loss: ", mloss / i_sentence)

    def save(self, filename):
        self.model.save(filename)

    def load(self, filename):
        self.model.populate(filename)


def _dict_to_obj(dic, name='Object'):
    """Return an object form of a dictionary."""
    return namedtuple(name, dic.keys())(*dic.values())



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from nlp_architect import LIBRARY_ROOT
from nlp_architect.data.cdc_resources.relations.relation_types_enums import \
    WikipediaSearchMethod, OnlineOROfflineMethod, EmbeddingMethod


class SievesResources(object):
    def __init__(self):
        """Cross Document co-reference sieve system resources configuration class"""

        self.__eval_output_dir = str(LIBRARY_ROOT / 'datasets' / 'cdc' / 'test_predict')

        self.__elastic_index = 'enwiki_v2'
        self.__elastic_host = 'localhost'
        self.__elastic_port = 9200
        self.__wiki_folder = str(LIBRARY_ROOT / 'dumps'  'wikipedia')
        self.__wd_file = str(LIBRARY_ROOT / 'dump' / 'within_doc_core' / 'ecb_wd_coref_proc.json')
        self.__wn_folder = str(LIBRARY_ROOT / 'dump' / 'wordnet')
        self.__elmo_file = str(LIBRARY_ROOT / 'dump' / 'embedde' / 'ecb_all_with_stop_elmo.pickle')
        self.__glove_file = str(LIBRARY_ROOT / 'dump' / 'embedde' / 'ecb_all_embed_glove.pickle')
        self.__referent_dict_file = str(LIBRARY_ROOT / 'dataset' / 'coref.dict1.tsv')
        self.__vo_dict_file = str(LIBRARY_ROOT / 'dataset' / 'verbocean.unrefined.2004-05-20.txt')

        self.__wiki_search_method = WikipediaSearchMethod.ONLINE
        self.__wn_search_method = OnlineOROfflineMethod.ONLINE
        self.__embed_search_method = EmbeddingMethod.ELMO
        self.__referent_dict_method = OnlineOROfflineMethod.ONLINE
        self.__vo_search_method = OnlineOROfflineMethod.ONLINE

    @property
    def eval_output_dir(self) -> str:
        """
        The output dir of the evaluation files, here scorer file for cross doc coref spans will
        be saved
        """
        return self.__eval_output_dir

    @eval_output_dir.setter
    def eval_output_dir(self, eval_output_dir: str):
        self.__eval_output_dir = eval_output_dir

    @property
    def wiki_folder(self):
        """
        Location of Wikipedia mini data set file, #Required mini data set file location for Offline
        evaluation using Wikipedia sieve
        """
        return self.__wiki_folder

    @wiki_folder.setter
    def wiki_folder(self, wiki_folder: str):
        self.__wiki_folder = wiki_folder

    @property
    def wd_file(self):
        """
        Location of Within doc data set file, #Required when using Within doc sieve
        """
        return self.__wd_file

    @wd_file.setter
    def wd_file(self, wd_file: str):
        self.__wd_file = wd_file

    @property
    def wn_folder(self):
        """
        Location of WordNet mini data set file, #Required mini data set file location for Offline
        evaluation using WordNet sieve
        """
        return self.__wn_folder

    @wn_folder.setter
    def wn_folder(self, wn_folder: str):
        self.__wn_folder = wn_folder

    @property
    def glove_file(self):
        """
        Location of GloVe mini data set file, #Required mini data set file location for Offline
        evaluation using GloVe sieve
        """
        return self.__glove_file

    @glove_file.setter
    def glove_file(self, glove_file: str):
        self.__glove_file = glove_file

    @property
    def elmo_file(self):
        """
        Location of Elmo mini data set file, #Required mini data set file location for Offline
        evaluation using GloVe sieve
        """
        return self.__elmo_file

    @elmo_file.setter
    def elmo_file(self, elmo_file: str):
        self.__elmo_file = elmo_file

    @property
    def referent_dict_file(self):
        """
        Location of Referent dic data set file, #Required mini data set file for Offline
        evaluation or original file for Online evaluation using Referent Dict sieve
        """
        return self.__referent_dict_file

    @referent_dict_file.setter
    def referent_dict_file(self, referent_dict_file: str):
        self.__referent_dict_file = referent_dict_file

    @property
    def vo_dict_file(self):
        """
        Location of VerbOcean data set file, #Required mini data set file for Offline evaluation or
        original file for Online evaluation using VerbOcean sieve
        """
        return self.__vo_dict_file

    @vo_dict_file.setter
    def vo_dict_file(self, vo_dict_file: str):
        self.__vo_dict_file = vo_dict_file

    @property
    def elastic_index(self):
        """
        Elastic index name, #Required when using Elastic evaluation using Wikipedia sieve
        """
        return self.__elastic_index

    @elastic_index.setter
    def elastic_index(self, elastic_index: str):
        self.__elastic_index = elastic_index

    @property
    def elastic_host(self):
        """
        Elastic host, #Required when using Elastic evaluation using Wikipedia sieve
        """
        return self.__elastic_host

    @elastic_host.setter
    def elastic_host(self, elastic_host: str):
        self.__elastic_host = elastic_host

    @property
    def elastic_port(self):
        """
        Elastic port number, #Required when using Elastic evaluation using Wikipedia sieve
        """
        return self.__elastic_port

    @elastic_port.setter
    def elastic_port(self, elastic_port: str):
        self.__elastic_port = elastic_port

    @property
    def wiki_search_method(self):
        """
        Wikipedia search method type, one of: WikipediaSearchMethod.ONLINE,
        WikipediaSearchMethod.OFFLINE, WikipediaSearchMethod.ELASTIC
        """
        return self.__wiki_search_method

    @wiki_search_method.setter
    def wiki_search_method(self, wiki_search_method: str):
        self.__wiki_search_method = wiki_search_method

    @property
    def wn_search_method(self):
        """
        Wordnet search method type, one of: OnlineOROfflineMethod.ONLINE,
        OnlineOROfflineMethod.OFFLINE
        """
        return self.__wn_search_method

    @wn_search_method.setter
    def wn_search_method(self, wn_search_method: str):
        self.__wn_search_method = wn_search_method

    @property
    def embed_search_method(self):
        """
        Wordnet search method type, one of: EmbeddingMethod.GLOVE, EmbeddingMethod.GLOVE_OFFLINE,
        EmbeddingMethod.ELMO, EmbeddingMethod.ELMO_OFFLINE
        """
        return self.__embed_search_method

    @embed_search_method.setter
    def embed_search_method(self, embed_search_method: str):
        self.__embed_search_method = embed_search_method

    @property
    def referent_dict_method(self):
        """
        Referent Dict search method type, one of: OnlineOROfflineMethod.ONLINE,
        OnlineOROfflineMethod.OFFLINE
        """
        return self.__referent_dict_method

    @referent_dict_method.setter
    def referent_dict_method(self, referent_dict_method: str):
        self.__referent_dict_method = referent_dict_method

    @property
    def vo_search_method(self):
        """
        VerbOcean search method type, one of: OnlineOROfflineMethod.ONLINE,
        OnlineOROfflineMethod.OFFLINE
        """
        return self.__vo_search_method

    @vo_search_method.setter
    def vo_search_method(self, vo_search_method: str):
        self.__vo_search_method = vo_search_method



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from nlp_architect.models.absa.inference.data_types import Polarity
from nlp_architect.models.absa.train.data_types import CandidateTerm, \
    RelCategory, DepRelationTerm, POS


def rule_1(dep_rel, gov_entry, dep_entry, text):
    """Extract term if rule 1 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        gov_entry (DicEntrySentiment): Governor opinion entry.
        dep_entry (DicEntrySentiment): Dependant opinion entry.
        text (str): Sentence text.
    """
    candidate = None
    (anchor_entry, anchor, related) = (gov_entry, dep_rel.gov, dep_rel.dep) \
        if gov_entry else (dep_entry, dep_rel.dep, dep_rel.gov)

    if related.norm_pos == POS.ADJ and dep_rel.rel.startswith('conj'):
        polarity = anchor_entry.polarity
        candidate = CandidateTerm(related, anchor, text, polarity)
    return candidate


def rule_2(dep_rel, relation_list, dep_entry, text):
    """Extract term if rule 2 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        relation_list (list of DepRelation): Generic relations between all tokens.
        dep_entry (OpinionTerm): Dependent token.
        text (str): Sentence text.
    """
    candidate = None
    for curr_rt in relation_list:
        if (curr_rt.gov, curr_rt.rel, curr_rt.dep.norm_pos) == \
                (dep_rel.gov, dep_rel.rel, POS.ADJ) and curr_rt.dep != dep_rel.dep:
            candidate = CandidateTerm(curr_rt.dep, dep_rel.dep, text, dep_entry.polarity)
    return candidate


def rule_3(dep_rel, relation_list, text):
    """Extract term if rule 3 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        relation_list (list of DepRelation): Generic relations between all tokens.
        text (str): Sentence text.
    """
    candidate = None
    if dep_rel.gov.norm_pos == POS.NN and is_subj_obj_or_mod(dep_rel):
        aspect = expand_aspect(dep_rel.gov, relation_list)
        candidate = CandidateTerm(aspect, dep_rel.dep, text, Polarity.UNK)
    return candidate


def rule_4(dep_rel, relation_list, text):
    """Extract term if rule 4 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        relation_list (list of DepRelation): Generic relations between all tokens.
        relation between tokens
        text (str): Sentence text.
    """
    candidate = None
    for curr_rt in relation_list:
        if curr_rt.gov == dep_rel.gov and curr_rt.dep != dep_rel.dep and \
                curr_rt.dep.norm_pos == POS.NN and is_subj_obj_or_mod(curr_rt) and \
                is_subj_obj_or_mod(dep_rel):
            aspect = expand_aspect(curr_rt.dep, relation_list)
            candidate = CandidateTerm(aspect, dep_rel.dep, text, Polarity.UNK)
    return candidate


def rule_5(dep_rel, text):
    """Extract term if rule 5 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        text (str): Sentence text.
    """
    candidate = None
    if is_subj_obj_or_mod(dep_rel) and dep_rel.dep.norm_pos == POS.ADJ:
        return CandidateTerm(dep_rel.dep, dep_rel.gov, text, Polarity.UNK)
    return candidate


def rule_6(dep_rel, relation_list, text):
    """Extract term if rule 6 applies.

    Args:
        dep_rel (DepRelation): Dependency relation.
        relation_list (list of DepRelation): Generic relations between all tokens.
        text (str): Sentence text.
    """
    candidate = None
    if dep_rel.rel in ('conj_and', 'conj_but'):
        aspect = expand_aspect(dep_rel.dep, relation_list)
        candidate = CandidateTerm(aspect, dep_rel.gov, text, Polarity.UNK)
    return candidate


def is_subj_obj_or_mod(rt):
    return any(rt.rel in cat.value for cat in (RelCategory.SUBJ, RelCategory.OBJ, RelCategory.MOD))


def expand_aspect(in_aspect_token, relation_list):
    """Expand aspect by Looking for a noun word that it's gov is the aspect. if it has (noun)
    compound relation add it to aspect."""
    aspect = DepRelationTerm(text=in_aspect_token.text, lemma=in_aspect_token.lemma,
                             pos=in_aspect_token.pos, ner=in_aspect_token.ner,
                             idx=in_aspect_token.idx)
    for rel in relation_list:
        if (rel.rel == 'compound') and (rel.gov.idx == aspect.idx):
            diff_positive = aspect.idx - len(rel.dep.text) - 1 - rel.dep.idx
            diff_negative = rel.dep.idx - len(aspect.text) - 1 - aspect.idx
            if diff_positive == 0:
                aspect.text = rel.dep.text + ' ' + aspect.text
                aspect.lemma = rel.dep.text + ' ' + aspect.lemma
                aspect.idx = rel.dep.idx
            if diff_negative == 0:
                aspect.text = aspect.text + ' ' + rel.dep.text
                aspect.lemma = aspect.lemma + ' ' + rel.dep.lemma
    return aspect



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import absolute_import, division, print_function, unicode_literals

import tensorflow as tf

from nlp_architect.contrib.tensorflow.python.keras.layers.crf import CRF
from nlp_architect.contrib.tensorflow.python.keras.utils.layer_utils import load_model, save_model


class NERCRF(object):
    """
    Bi-LSTM NER model with CRF classification layer (tf.keras model)

    Args:
        use_cudnn (bool, optional): use cudnn LSTM cells
    """

    def __init__(self, use_cudnn=False):
        self.model = None
        self.word_length = None
        self.target_label_dims = None
        self.word_vocab_size = None
        self.char_vocab_size = None
        self.word_embedding_dims = None
        self.char_embedding_dims = None
        self.tagger_lstm_dims = None
        self.dropout = None
        self.use_cudnn = use_cudnn

    def build(self,
              word_length,
              target_label_dims,
              word_vocab_size,
              char_vocab_size,
              word_embedding_dims=100,
              char_embedding_dims=16,
              tagger_lstm_dims=200,
              dropout=0.5):
        """
        Build a NERCRF model

        Args:
            word_length (int): max word length in characters
            target_label_dims (int): number of entity labels (for classification)
            word_vocab_size (int): word vocabulary size
            char_vocab_size (int): character vocabulary size
            word_embedding_dims (int): word embedding dimensions
            char_embedding_dims (int): character embedding dimensions
            tagger_lstm_dims (int): word tagger LSTM output dimensions
            dropout (float): dropout rate
        """
        self.word_length = word_length
        self.target_label_dims = target_label_dims
        self.word_vocab_size = word_vocab_size
        self.char_vocab_size = char_vocab_size
        self.word_embedding_dims = word_embedding_dims
        self.char_embedding_dims = char_embedding_dims
        self.tagger_lstm_dims = tagger_lstm_dims
        self.dropout = dropout

        # build word input
        words_input = tf.keras.layers.Input(shape=(None,), name='words_input')
        embedding_layer = tf.keras.layers.Embedding(self.word_vocab_size, self.word_embedding_dims,
                                                    name='word_embedding')
        word_embeddings = embedding_layer(words_input)

        # create word character embeddings
        word_chars_input = tf.keras.layers.Input(shape=(None, self.word_length),
                                                 name='word_chars_input')
        char_embedding_layer = tf.keras.layers.Embedding(self.char_vocab_size,
                                                         self.char_embedding_dims,
                                                         name='char_embedding')(word_chars_input)
        char_embeddings = \
            tf.keras.layers.TimeDistributed(
                tf.keras.layers.Conv1D(128, 3, padding='same', activation='relu'))(
                char_embedding_layer)
        char_embeddings = tf.keras.layers.TimeDistributed(tf.keras.layers.GlobalMaxPooling1D())(
            char_embeddings)

        # create the final feature vectors
        features = tf.keras.layers.concatenate([word_embeddings, char_embeddings], axis=-1)

        # encode using a bi-LSTM
        features = tf.keras.layers.Dropout(self.dropout)(features)
        bilstm = tf.keras.layers.Bidirectional(self._rnn_cell(self.tagger_lstm_dims,
                                                              return_sequences=True))(features)
        bilstm = tf.keras.layers.Bidirectional(self._rnn_cell(self.tagger_lstm_dims,
                                                              return_sequences=True))(bilstm)
        bilstm = tf.keras.layers.Dropout(self.dropout)(bilstm)
        bilstm = tf.keras.layers.Dense(self.target_label_dims)(bilstm)

        inputs = [words_input, word_chars_input]

        sequence_lengths = tf.keras.layers.Input(shape=(1,), dtype='int32', name='seq_lens')
        inputs.append(sequence_lengths)
        crf = CRF(self.target_label_dims, name='ner_crf')
        predictions = crf(inputs=bilstm, sequence_lengths=sequence_lengths)

        # compile the model
        model = tf.keras.Model(inputs=inputs,
                               outputs=predictions)
        model.compile(loss={'ner_crf': crf.loss},
                      optimizer=tf.keras.optimizers.Adam(0.001, clipnorm=5.))

        self.model = model

    def _rnn_cell(self, units, **kwargs):
        if self.use_cudnn:
            rnn_cell = tf.keras.layers.CuDNNLSTM(units, **kwargs)
        else:
            rnn_cell = tf.keras.layers.LSTM(units, **kwargs)
        return rnn_cell

    def load_embedding_weights(self, weights):
        """
        Load word embedding weights into the model embedding layer

        Args:
            weights (numpy.ndarray): 2D matrix of word weights
        """
        assert self.model is not None, 'Cannot assign weights, apply build() before trying to ' \
                                       'loading embedding weights '
        emb_layer = self.model.get_layer(name='word_embedding')
        assert emb_layer.output_dim == weights.shape[1], 'embedding vectors shape mismatch'
        emb_layer.set_weights([weights])

    def fit(self, x, y, epochs=1, batch_size=1, callbacks=None, validation=None):
        """
        Train a model given input samples and target labels.

        Args:
            x (numpy.ndarray or :obj:`numpy.ndarray`): input samples
            y (numpy.ndarray): input sample labels
            epochs (:obj:`int`, optional): number of epochs to train
            batch_size (:obj:`int`, optional): batch size
            callbacks(:obj:`Callback`, optional): Keras compatible callbacks
            validation(:obj:`list` of :obj:`numpy.ndarray`, optional): optional validation data
                to be evaluated when training
        """
        assert self.model, 'Model was not initialized'
        self.model.fit(x, y, epochs=epochs, batch_size=batch_size, shuffle=True,
                       validation_data=validation,
                       callbacks=callbacks)

    def predict(self, x, batch_size=1):
        """
        Get the prediction of the model on given input

        Args:
            x (numpy.ndarray or :obj:`numpy.ndarray`): input samples
            batch_size (:obj:`int`, optional): batch size

        Returns:
            numpy.ndarray: predicted values by the model
        """
        assert self.model, 'Model was not initialized'
        return self.model.predict(x, batch_size=batch_size)

    def save(self, path):
        """
        Save model to path

        Args:
            path (str): path to save model weights
        """
        topology = {k: v for k, v in self.__dict__.items()}
        topology.pop('model')
        topology.pop('use_cudnn')
        save_model(self.model, topology, path)

    def load(self, path):
        """
        Load model weights

        Args:
            path (str): path to load model from
        """
        load_model(path, self)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""Generally useful utility functions."""
from __future__ import print_function

import codecs
import collections
import json
import math
import os
import sys
import time
from distutils import version

import tensorflow as tf


def check_tensorflow_version():
    min_tf_version = "1.4.0-dev20171024"
    if (version.LooseVersion(tf.__version__) < version.LooseVersion(min_tf_version)):
        raise EnvironmentError("Tensorflow version must >= %s" % min_tf_version)


def safe_exp(value):
    """Exponentiation with catching of overflow error."""
    try:
        ans = math.exp(value)
    except OverflowError:
        ans = float("inf")
    return ans


def print_time(s, start_time):
    """Take a start time, print elapsed duration, and return a new time."""
    print("%s, time %ds, %s." % (s, (time.time() - start_time), time.ctime()))
    sys.stdout.flush()
    return time.time()


def print_out(s, f=None, new_line=True):
    """Similar to print but with support to flush and output to a file."""
    if isinstance(s, bytes):
        s = s.decode("utf-8")

    if f:
        f.write(s.encode("utf-8"))
        if new_line:
            f.write(b"\n")

    # stdout
    out_s = s.encode("utf-8")
    if not isinstance(out_s, str):
        out_s = out_s.decode("utf-8")
    print(out_s, end="", file=sys.stdout)

    if new_line:
        sys.stdout.write("\n")
    sys.stdout.flush()


def print_hparams(hparams, skip_patterns=None, header=None):
    """Print hparams, can skip keys based on pattern."""
    if header:
        print_out("%s" % header)
    values = hparams.values()
    for key in sorted(values.keys()):
        if not skip_patterns or all(
                [skip_pattern not in key for skip_pattern in skip_patterns]):
            print_out("  %s=%s" % (key, str(values[key])))


def load_hparams(model_dir):
    """Load hparams from an existing model directory."""
    hparams_file = os.path.join(model_dir, "hparams")
    if tf.gfile.Exists(hparams_file):
        print_out("# Loading hparams from %s" % hparams_file)
        with codecs.getreader("utf-8")(tf.gfile.GFile(hparams_file, "rb")) as f:
            try:
                hparams_values = json.load(f)
                hparams = tf.contrib.training.HParams(**hparams_values)
            except ValueError:
                print_out("  can't load hparams file")
                return None
        return hparams
    else:
        return None


def maybe_parse_standard_hparams(hparams, hparams_path):
    """Override hparams values with existing standard hparams config."""
    if hparams_path and tf.gfile.Exists(hparams_path):
        print_out("# Loading standard hparams from %s" % hparams_path)
        with codecs.getreader("utf-8")(tf.gfile.GFile(hparams_path, "rb")) as f:
            hparams.parse_json(f.read())
    return hparams


def save_hparams(out_dir, hparams):
    """Save hparams."""
    hparams_file = os.path.join(out_dir, "hparams")
    print_out("  saving hparams to %s" % hparams_file)
    with codecs.getwriter("utf-8")(tf.gfile.GFile(hparams_file, "wb")) as f:
        f.write(hparams.to_json(indent=4, sort_keys=True))


def debug_tensor(s, msg=None, summarize=10):
    """Print the shape and value of a tensor at test time. Return a new tensor."""
    if not msg:
        msg = s.name
    return tf.Print(s, [tf.shape(s), s], msg + " ", summarize=summarize)


def add_summary(summary_writer, global_step, tag, value):
    """Add a new summary to the current summary_writer.
  Useful to log things that are not part of the training graph, e.g., tag=BLEU.
  """
    summary = tf.Summary(value=[tf.Summary.Value(tag=tag, simple_value=value)])
    summary_writer.add_summary(summary, global_step)


def get_config_proto(log_device_placement=False, allow_soft_placement=True,
                     num_intra_threads=0, num_inter_threads=0):
    # GPU options:
    # https://www.tensorflow.org/versions/r0.10/how_tos/using_gpu/index.html
    config_proto = tf.ConfigProto(
        log_device_placement=log_device_placement,
        allow_soft_placement=allow_soft_placement)
    config_proto.gpu_options.allow_growth = True

    # CPU threads options
    if num_intra_threads:
        config_proto.intra_op_parallelism_threads = num_intra_threads
    if num_inter_threads:
        config_proto.inter_op_parallelism_threads = num_inter_threads

    return config_proto


def format_text(words):
    """Convert a sequence words into sentence."""
    if (not hasattr(words, "__len__")  # for numpy array
            and not isinstance(words, collections.Iterable)):
        words = [words]
    return b" ".join(words)


def format_bpe_text(symbols, delimiter=b"@@"):
    """Convert a sequence of bpe words into sentence."""
    words = []
    word = b""
    if isinstance(symbols, str):
        symbols = symbols.encode()
    delimiter_len = len(delimiter)
    for symbol in symbols:
        if len(symbol) >= delimiter_len and symbol[-delimiter_len:] == delimiter:
            word += symbol[:-delimiter_len]
        else:  # end of a word
            word += symbol
            words.append(word)
            word = b""
    return b" ".join(words)


def format_spm_text(symbols):
    """Decode a text in SPM (https://github.com/google/sentencepiece) format."""
    return u"".join(format_text(symbols).decode("utf-8").split()).replace(
        u"\u2581", u" ").strip().encode("utf-8")



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# pylint: disable=no-name-in-module
import itertools
import math

import numpy as np
from spacy.tokens.doc import Doc
from spacy.tokens.span import Span
from spacy.tokens.token import Token
from wordfreq import zipf_frequency


class TextSpanScoring:
    """
    Text spans scoring class.
    Contains misc scoring algorithms for scoring text fragments extracted
    from a corpus.

    Arguments:
        documents(list): List of spaCy documents.
        spans(list[list]): List of spaCy spans representing noun phrases of documents
            document.
    """

    def __init__(self, documents, spans, min_tf=1):
        assert len(documents) == len(spans)
        self._documents = documents
        self._doc_text_spans = spans
        self.index = CorpusIndex(documents, spans)
        assert min_tf > 0, 'min_tf must be > 0'
        if min_tf > 1:
            self.re_index_min_tf(min_tf)

    def re_index_min_tf(self, tf):
        filtered_doc_text_spans = []
        for d in self.doc_text_spans:
            filtered_doc_phrases = [p for p in d if self.index.tf(p) >= tf]
            filtered_doc_text_spans.append(filtered_doc_phrases)
        self._doc_text_spans = filtered_doc_text_spans
        self.index = CorpusIndex(self.documents, self.doc_text_spans)

    @property
    def documents(self):
        return self._documents

    @property
    def doc_text_spans(self):
        return self._doc_text_spans

    def get_tfidf_scores(self, group_similar_spans=True):
        """
        Get TF-IDF scores of spans
        return a list of spans sorted by desc order of importance
        span score = TF (global) * (1 + log_n(DF/N))
        """
        phrases_and_scores = {}
        num_of_docs = len(self.documents)
        for _, noun_phrases in zip(self.documents, self.doc_text_spans):
            for p in noun_phrases:
                if p not in phrases_and_scores:
                    tf = self.index.tf(p)
                    df = self.index.df(p)
                    phrases_and_scores[p] = \
                        (tf + 1) * math.log(1 + num_of_docs / df)
        if len(phrases_and_scores) > 0:
            return self._maybe_group_and_sort(group_similar_spans,
                                              phrases_and_scores)
        return []

    def get_freq_scores(self, group_similar_spans=True):
        phrases_and_scores = {}
        for _, noun_phrases in zip(self.documents, self.doc_text_spans):
            for p in noun_phrases:
                if p not in phrases_and_scores:
                    phrases_and_scores[p] = zipf_frequency(p.text, 'en')
        return self._maybe_group_and_sort(group_similar_spans,
                                          phrases_and_scores)

    def group_spans(self, phrases):
        pid_phrase_scores = [{'k': self.index.get_pid(p),
                              'v': (p, s)}
                             for p, s in phrases.items()]
        phrase_groups = []
        for _, group in itertools.groupby(
                sorted(pid_phrase_scores, key=lambda x: x['k']),
                lambda x: x['k']):
            _group = list(group)
            phrases = {g['v'][0].text for g in _group}
            score = _group[0]['v'][1]
            phrase_groups.append((sorted(phrases), score))
        return phrase_groups

    def _maybe_group_and_sort(self, is_group, phrases_dict):
        phrase_groups = phrases_dict.items()
        if is_group:
            phrase_groups = self.group_spans(phrases_dict)
        return sorted(phrase_groups, key=lambda x: x[1], reverse=True)

    @staticmethod
    def normalize_minmax(phrases_list, invert=False):
        _, scores = list(zip(*phrases_list))
        max_score = max(scores)
        min_score = min(scores)
        norm_list = []
        for p, s in phrases_list:
            if max_score - min_score > 0:
                new_score = (s - min_score) / (max_score - min_score)
            else:
                new_score = 0
            norm_list.append([p, new_score])
        if invert:
            for e in norm_list:
                e[1] = 1.0 - e[1]
        return norm_list

    @staticmethod
    def normalize_l2(phrases_list):
        phrases, scores = list(zip(*phrases_list))
        scores = scores / np.linalg.norm(scores)
        return list(zip(phrases, scores.tolist()))

    @staticmethod
    def interpolate_scores(phrase_lists, weights=None):
        if weights is None:
            weights = [1.0 / len(phrase_lists)]
        else:
            assert len(weights) == len(phrase_lists)

        list_sizes = [len(l) for l in phrase_lists]
        for l in list_sizes:
            assert len(phrase_lists[0]) == l, 'list sizes not equal'

        phrase_list_dicts = []
        for lst in phrase_lists:
            phrase_list_dicts.append({tuple(k): v for k, v in lst})

        phrases = phrase_list_dicts[0].keys()
        interp_scores = {}
        for p in phrases:
            interp_scores[p] = 0.0
            for ref, w in zip(phrase_list_dicts, weights):
                interp_scores[p] += ref[p] * w
        return sorted(interp_scores.items(), key=lambda x: x[1], reverse=True)

    def get_cvalue_scores(self, group_similar_spans=True):
        phrases_text = [p for ps in self.index.pid_to_spans.values() for p in ps]
        phrase_scores = {}
        for phrase in phrases_text:
            sub_phrase_pids = None
            for w in phrase:
                sub_pids = self.index.get_subphrases_of_word(w)
                if sub_phrase_pids is None:
                    sub_phrase_pids = sub_pids
                else:
                    sub_phrase_pids = sub_phrase_pids.intersection(sub_pids)
            sub_phrase_pids.discard(self.index.get_pid(phrase))

            if len(sub_phrase_pids) > 0:
                score = math.log2(1 + len(phrase)) * (
                    self.index.tf(phrase) - 1.0 / len(sub_phrase_pids) * sum(
                        [self.index.tf(list(self.index.get_phrase(p))[0])
                         for p in sub_phrase_pids]))
            else:
                score = math.log2(1 + len(phrase)) * self.index.tf(phrase)
            phrase_scores[phrase] = score
        for phrase in phrase_scores:
            score = [phrase_scores[p] for p in
                     self.index.get_phrase(self.index.get_pid(phrase))]
            phrase_scores[phrase] = sum(score) / len(score)
        return self._maybe_group_and_sort(group_similar_spans,
                                          phrase_scores)

    @staticmethod
    def multiply_scores(phrase_lists):
        phrase_list_dicts = []
        for lst in phrase_lists:
            phrase_list_dicts.append({tuple(k): v for k, v in lst})
        phrases = phrase_list_dicts[0].keys()
        interp_scores = {}
        for p in phrases:
            interp_scores[p] = 1.0
            for ref in phrase_list_dicts:
                interp_scores[p] *= ref[p]
        return sorted(interp_scores.items(), key=lambda x: x[1], reverse=True)


class CorpusIndex:
    """
    Text span index class.
    Holds TF and DF values per span. Text spans are normalized and similar
    spans are mapped to the same TF DF values.
    """

    def __init__(self, documents: list, spans: list):
        self.df_index = {}  # span to DF
        self.tf_index = {}  # span, doc to TF
        self.pid_to_spans = {}  # id to spans (normalized types)
        self.word_to_phrase_index = {}
        self.documents = set()
        for d, phrases in zip(documents, spans):
            doc_id = CorpusIndex.get_docid(d)
            self.documents.add(doc_id)
            for phrase in phrases:
                pid = CorpusIndex.get_pid(phrase)
                # add to norm phrases dict
                if pid not in self.pid_to_spans:
                    self.pid_to_spans[pid] = set()
                self.pid_to_spans[pid].add(phrase)

                # add to df index
                if self.df_index.get(pid, None) is None:
                    self.df_index[pid] = {doc_id}
                else:
                    self.df_index[pid].add(doc_id)

                # add to tf index
                if self.tf_index.get(pid, None) is None:
                    self.tf_index[pid] = 1
                else:
                    self.tf_index[pid] += 1

                for word in phrase:
                    wid = self.get_wid(word)
                    if wid not in self.word_to_phrase_index:
                        self.word_to_phrase_index[wid] = set()
                    self.word_to_phrase_index[wid].add(pid)

    def get_phrase(self, pid):
        return self.pid_to_spans.get(pid)

    def get_subphrases_of_word(self, w):
        wid = self.get_wid(w)
        return self.word_to_phrase_index.get(wid)

    @staticmethod
    def get_wid(w: Token):
        """
        get word id
        """
        return CorpusIndex.hash_func(w.text)

    @staticmethod
    def get_pid(p: Span):
        """
        get phrase id
        """
        return CorpusIndex.hash_func(p.lemma_)

    @staticmethod
    def get_docid(d: Doc):
        """
        get doc id
        """
        return CorpusIndex.hash_func(d)

    @staticmethod
    def hash_func(x):
        return hash(x)

    def tf(self, phrase):
        """
        Get TF of phrase in doc
        """
        pid = CorpusIndex.get_pid(phrase)
        if self.tf_index.get(pid, None) is not None:
            return self.tf_index.get(pid)
        return 0

    def df(self, phrase):
        """
        Get DF of phrase in corpus
        """
        pid = CorpusIndex.get_pid(phrase)
        if self.df_index.get(pid, None) is not None:
            return len(self.df_index[pid])
        return 0



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import json
import os

from nlp_architect.models.bist import utils
from nlp_architect.models.bist.mstlstm import MSTParserLSTM
from nlp_architect.models.bist.utils import get_options_dict
from nlp_architect.utils.io import validate, validate_existing_filepath


class BISTModel(object):
    """
    BIST parser model class.
    This class handles training, prediction, loading and saving of a BIST parser model.
    After the model is initialized, it accepts a CoNLL formatted dataset as input, and learns to
    output dependencies for new input.

    Args:
        activation (str, optional): Activation function to use.
        lstm_layers (int, optional): Number of LSTM layers to use.
        lstm_dims (int, optional): Number of LSTM dimensions to use.
        pos_dims (int, optional): Number of part-of-speech embedding dimensions to use.

    Attributes:
        model (MSTParserLSTM): The underlying LSTM model.
        params (tuple): Additional parameters and resources for the model.
        options (dict): User model options.
    """

    def __init__(self, activation='tanh', lstm_layers=2, lstm_dims=125, pos_dims=25):
        validate((activation, str), (lstm_layers, int, 0, None), (lstm_dims, int, 0, 1000),
                 (pos_dims, int, 0, 1000))
        self.options = get_options_dict(activation, lstm_dims, lstm_layers, pos_dims)
        self.params = None
        self.model = None

    def fit(self, dataset, epochs=10, dev=None):
        """
        Trains a BIST model on an annotated dataset in CoNLL file format.

        Args:
            dataset (str): Path to input dataset for training, formatted in CoNLL/U format.
            epochs (int, optional): Number of learning iterations.
            dev (str, optional): Path to development dataset for conducting evaluations.
        """
        if dev:
            dev = validate_existing_filepath(dev)
        dataset = validate_existing_filepath(dataset)
        validate((epochs, int, 0, None))

        print('\nRunning fit on ' + dataset + '...\n')
        words, w2i, pos, rels = utils.vocab(dataset)
        self.params = words, w2i, pos, rels, self.options
        self.model = MSTParserLSTM(*self.params)

        for epoch in range(epochs):
            print('Starting epoch', epoch + 1)
            self.model.train(dataset)
            if dev:
                ext = dev.rindex('.')
                res_path = dev[:ext] + '_epoch_' + str(epoch + 1) + '_pred' + dev[ext:]
                utils.write_conll(res_path, self.model.predict(dev))
                utils.run_eval(dev, res_path)

    def predict(self, dataset, evaluate=False):
        """
        Runs inference with the BIST model on a dataset in CoNLL file format.

        Args:
            dataset (str): Path to input CoNLL file.
            evaluate (bool, optional): Write prediction and evaluation files to dataset's folder.
        Returns:
            res (list of list of ConllEntry): The list of input sentences with predicted
            dependencies attached.
        """
        dataset = validate_existing_filepath(dataset)
        validate((evaluate, bool))

        print('\nRunning predict on ' + dataset + '...\n')
        res = list(self.model.predict(conll_path=dataset))
        if evaluate:
            ext = dataset.rindex('.')
            pred_path = dataset[:ext] + '_pred' + dataset[ext:]
            utils.write_conll(pred_path, res)
            utils.run_eval(dataset, pred_path)
        return res

    def predict_conll(self, dataset):
        """
        Runs inference with the BIST model on a dataset in CoNLL object format.

        Args:
            dataset (list of list of ConllEntry): Input in the form of ConllEntry objects.
        Returns:
            res (list of list of ConllEntry): The list of input sentences with predicted
            dependencies attached.
        """
        res = None
        if hasattr(dataset, '__iter__'):
            res = list(self.model.predict(conll=dataset))
        return res

    def load(self, path):
        """Loads and initializes a BIST model from file."""
        with open(path.parent / 'params.json') as file:
            self.params = json.load(file)
        self.model = MSTParserLSTM(*self.params)
        self.model.model.populate(str(path))

    def save(self, path):
        """Saves the BIST model to file."""
        print("Saving")
        with open(os.path.join(os.path.dirname(path), 'params.json'), 'w') as file:
            json.dump(self.params, file)
        self.model.model.save(path)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import os
import logging
import re

from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page import WikipediaPage
from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page_extracted_relations import \
    WikipediaPageExtractedRelations
from nlp_architect.data.cdc_resources.wikipedia.wiki_search_page_result import \
    WikipediaSearchPageResult
from nlp_architect.utils.text import SpacyInstance

os.environ['PYWIKIBOT_NO_USER_CONFIG'] = '1'

DISAMBIGUATE_PAGE = ['wikimedia disambiguation page', 'wikipedia disambiguation page']
NAME_DESCRIPTIONS = ['given name', 'first name', 'family name']

logger = logging.getLogger(__name__)


class WikiOnline(object):
    def __init__(self):
        import pywikibot
        self.spacy = SpacyInstance()
        self.pywikibot = pywikibot
        self.cache = dict()
        self.site = pywikibot.Site('en', 'wikipedia')  # The site we want to run our bot on

    def get_pages(self, phrase):
        if phrase in self.cache:
            return self.cache[phrase]

        ret_pages = set()
        word_clean = phrase.replace('-', ' ')
        word_lower = word_clean.lower()
        word_upper = word_clean.upper()
        word_title = word_clean.title()
        words_set = {phrase, word_clean, word_lower, word_upper, word_title}
        for appr in words_set:
            try:
                page_result = self.get_page_redirect(appr)
                if page_result.pageid != 0:
                    full_page = self.get_wiki_page_with_items(phrase, page_result)
                    ret_pages.add(WikipediaSearchPageResult(appr, full_page))
            except Exception as e:
                logger.error(e)

        self.cache[phrase] = ret_pages
        return ret_pages

    # pylint: disable=protected-access
    def get_wiki_page_with_items(self, phrase, page):
        item = self.get_wiki_page_item(page)
        pageid = page.pageid
        aliases = self.get_aliases(item)
        description = self.get_description(item)
        text = page.text
        page_title = page._link._title

        relations = WikipediaPageExtractedRelations()
        relations.is_disambiguation = self.is_disambiguation_page(item)
        relations.is_part_name = self.is_name_description(text, item, relations.is_disambiguation)
        relations.aliases = aliases
        relations.be_comp, relations.be_comp_norm = self.extract_be_comp(text)
        relations.extract_relations_from_text_v0(text)

        ret_page = WikipediaPage(phrase, None, page_title, None, 0, pageid, description, relations)

        logger.debug('Page: {}. Extracted successfully'.format(ret_page))

        return ret_page

    def get_wiki_page_item(self, page):
        if page is not None:
            try:
                item = self.pywikibot.ItemPage.fromPage(
                    page)  # this can be used for any page object
                item.get()  # need to call it to access any data.
                return item
            except (self.pywikibot.NoPage, AttributeError, TypeError, NameError):
                pass
        return None

    def get_page_redirect(self, word):
        page = self.pywikibot.Page(self.site, word)
        if page.pageid != 0 and page.isRedirectPage():
            return page.getRedirectTarget()
        return page

    @staticmethod
    def get_aliases(item):
        if item is not None and item.aliases is not None:
            if 'en' in item.aliases:
                aliases = item.aliases['en']
                return aliases

        return None

    @staticmethod
    def get_description(item):
        description = {}
        if item is not None:
            item_desc = item.get()
            if 'desctiptions' in item_desc and 'en' in item_desc['descriptions']:
                dict([("age", 25)])
                description['descriptions'] = dict([('en', item_desc['descriptions']['en'])])

        return description

    @staticmethod
    def is_disambiguation_page(item):
        if item is not None:
            dic = item.get()
            if dic is not None and 'descriptions' in dic:
                desc = dic['descriptions']
                if desc is not None and 'en' in desc:
                    return desc['en'].lower()in DISAMBIGUATE_PAGE

        return False

    @staticmethod
    def is_name_description(text, item, is_disambiguation):
        if item is not None:
            if is_disambiguation:
                if WikipediaPageExtractedRelations.is_name_part(text):
                    return True
            else:
                dic = item.get()
                if dic is not None and 'descriptions' in dic:
                    desc = dic['descriptions']
                    if desc is not None and 'en' in desc:
                        if [s for s in NAME_DESCRIPTIONS if s in desc['en'].lower()]:
                            return True
        return False

    # pylint: disable=no-else-return
    def extract_be_comp(self, text):
        first_sentence_start_index = text.index("'''")
        if first_sentence_start_index >= 0:
            last_temp_index = text.find('\n', first_sentence_start_index)
        if last_temp_index == -1:
            last_temp_index = len(text)

        first_paragraph = text[first_sentence_start_index:last_temp_index]
        if WikiOnline.extract_be_a_index(first_paragraph) == -1 and last_temp_index != len(text):
            return self.extract_be_comp(text[last_temp_index:])
        elif last_temp_index == len(text):
            return None, None

        first_paragraph_clean = re.sub(r'\([^)]*\)', '', first_paragraph)
        first_paragraph_clean = re.sub(r'<[^>]*>', '', first_paragraph_clean)
        first_paragraph_clean = re.sub(r'{[^}]*}', '', first_paragraph_clean)
        first_paragraph_clean = re.sub(r'\[\[[^]]*\]\]', '', first_paragraph_clean)
        first_paragraph_clean = re.sub(r'[\']', '', first_paragraph_clean)
        first_paragraph_clean = re.sub(r'&nbsp;', ' ', first_paragraph_clean)

        return self.extract_be_comp_relations(first_paragraph_clean)

    # pylint: disable=not-callable
    def extract_be_comp_relations(self, first_paragraph):
        be_comp = set()
        be_comp_norm = set()
        if first_paragraph:
            doc = self.spacy.parser(first_paragraph)
            for token in doc:
                target = token.text
                target_lemma = token.lemma_
                relation = token.dep_
                governor = token.head.text
                governor_lemma = token.head.lemma_
                if relation == 'acl':
                    break
                if relation == 'punct' and target == '.':
                    break
                elif relation == 'cop':
                    be_comp.add(governor)
                    be_comp_norm.add(governor_lemma)
                elif relation == 'nsubj':
                    be_comp.add(target)
                    be_comp_norm.add(target_lemma)
                elif relation == 'dep':
                    be_comp.add(governor)
                    be_comp_norm.add(governor_lemma)
                elif relation == 'compound':
                    be_comp.add(target + ' ' + governor)
                    be_comp_norm.add(target_lemma + ' ' + governor_lemma)
                elif relation == 'amod':
                    be_comp.add(target + ' ' + governor)
                    be_comp_norm.add(target_lemma + ' ' + governor_lemma)
                elif relation in ['conj', 'appos']:
                    be_comp.add(target)
                    be_comp_norm.add(target_lemma)

        return be_comp, be_comp_norm

    @staticmethod
    def extract_be_a_index(sentence):
        result = None
        if 'is a' in sentence:
            result = sentence.index("is a")
        elif 'are a' in sentence:
            result = sentence.index("are a")
        elif 'was a' in sentence:
            result = sentence.index("was a")
        elif 'were a' in sentence:
            result = sentence.index("were a")
        elif 'be a' in sentence:
            result = sentence.index("be a")
        elif 'is the' in sentence:
            result = sentence.index("is the")
        elif 'are the' in sentence:
            result = sentence.index("are the")
        elif 'was the' in sentence:
            result = sentence.index("was the")
        elif 'were the' in sentence:
            result = sentence.index("were the")
        elif 'be the' in sentence:
            result = sentence.index("be the")

        return result



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import csv
import pickle
import numpy as np
import tensorflow
from os import PathLike


from nlp_architect.models.absa.utils import _read_generic_lex_for_similarity
from nlp_architect.models.absa import TRAIN_OUT, TRAIN_LEXICONS

from scipy.spatial.distance import cosine
from sklearn.model_selection import StratifiedKFold
# pylint: disable=import-error
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.models import Sequential, load_model

from nlp_architect.utils.embedding import load_word_embeddings


class RerankTerms(object):
    out_dir = TRAIN_OUT / 'output'
    model_dir = TRAIN_OUT / 'reranking_model'
    train_rerank_data_path = TRAIN_LEXICONS / 'RerankTrainingData.csv'
    PREDICTION_THRESHOLD = 0.7

    def __init__(self, vector_cache=True, rerank_model: PathLike = None,
                 emb_model_path: PathLike = None):
        # model and training params
        self.embeddings_len = 300
        self.activation_1 = 'relu'
        self.activation_2 = 'relu'
        self.activation_3 = 'sigmoid'
        self.loss = 'binary_crossentropy'
        self.optimizer = 'rmsprop'

        self.epochs_and_batch_size = [(10, 2)]
        self.seeds = [3]
        self.threshold = 0.5

        self.sim_lexicon = TRAIN_LEXICONS / 'RerankSentSimLex.csv'
        self.generic_lexicon = TRAIN_LEXICONS / 'GenericOpinionLex.csv'

        self.vector_cache = vector_cache
        self.word_vectors_dict = {}
        self.vectors_sim_dict = {}

        self.rerank_model_path = rerank_model
        self.emb_model_path = emb_model_path

        RerankTerms.out_dir.mkdir(parents=True, exist_ok=True)

        tensorflow.logging.set_verbosity(tensorflow.logging.ERROR)

    def calc_cosine_similarity(self, word_1, word_2, embedding_dict):
        """
        calculate cosine similarity scores between 2 terms

        Args:
            word_1 (str): 1st input word
            word_2 (str): 2nd input word
            embedding_dict (dict): embedding dictionary

        Returns:
            vectors_sim_dict[key] (float): similarity scores between the 2 input words

        """
        key = str(sorted([word_1, word_2]))

        if not self.vector_cache or key not in self.vectors_sim_dict:
            vector_1 = embedding_dict.get(word_1)
            vector_2 = embedding_dict.get(word_2)

            # check if both words have vectors
            if np.count_nonzero(vector_1) > 0 and np.count_nonzero(vector_2) > 0:
                sim_score = cosine(vector_1, vector_2)
            else:
                sim_score = None
            self.vectors_sim_dict[key] = sim_score

        return self.vectors_sim_dict[key]

    def calc_similarity_scores_for_all_terms(self, terms, generic_terms, embedding_dict):
        """
        calculate similarity scores between each term and each off the generic terms

        Args:
            terms: candidate terms
            generic_terms: generic opinion terms
            embedding_dict: embedding dictionary

        Returns:
            neg_all: similarity scores between each cand term and neg generic term
            pos_all: similarity scores between each cand term and pos generic term

        """
        print("\nComputing similarity scores...\n")

        neg_all = []
        pos_all = []

        for term in terms:
            polarity_sim_dic = {'NEG': [], 'POS': []}
            for generic_term, polarity in generic_terms.items():

                sim_score = self.calc_cosine_similarity(term, generic_term, embedding_dict)

                if sim_score is not None:
                    polarity_sim_dic[polarity].append(sim_score)
                else:
                    polarity_sim_dic[polarity].append(float(0))

            neg_all.append(polarity_sim_dic['NEG'])
            pos_all.append(polarity_sim_dic['POS'])

        return neg_all, pos_all

    @staticmethod
    def load_terms_and_polarities(filename):
        """
        load terms and polarities from file

        Args:
            filename: feature table file full path

        Returns:
            terms: candidate terms
            polarities: opinion polarity per term

        """
        print('Loading training data from {} ...'.format(filename))

        table = np.genfromtxt(filename, delimiter=',', skip_header=1, dtype=str)
        if table.size == 0:
            raise ValueError('Error: Term file is empty, no terms to re-rank.')

        try:
            terms = table[:, 1]
        except Exception as e:
            print("\n\nError converting str to float in training table: {}".format(e))

        polarities = table[:, 3].astype(str)

        if len(terms) != len(polarities):
            raise ValueError(
                'Count of opinion terms is different than the count of loaded polarities.')
        polarities = {terms[i]: polarities[i] for i in range(len(terms))}

        print(str(terms.shape[0]) + ' features loaded from CSV file')
        return terms, polarities

    @staticmethod
    def load_terms_and_y_labels(filename):
        """Load terms and Y labels from feature file.

        Args:
            filename: feature table file full path

        Returns:
            x: feature vector
            y: labels vector
            terms: candidate terms
            polarities: opinion polarity per term
        """
        print('Loading basic features from {} ...'.format(filename))

        table = np.genfromtxt(filename, delimiter=',', skip_header=1, dtype=str)
        if table.size == 0:
            raise ValueError('Error: Terms file is empty, no terms to re-rank.')

        try:
            terms = table[:, 1]
        except Exception as e:
            print("\n\nError converting str to float in training table: {}".format(e))

        y = table[:, 0].astype(int)
        polarities = None

        print(str(terms.shape[0]) + ' features loaded from CSV file')
        return y, terms, polarities

    @staticmethod
    def concat_sim_scores_and_features(x, neg_sim, pos_sim):
        """
        concatenate similarity scores to features

        Args:
            x: feature vector
            neg_sim: similarity scores between cand terms and neg opinion terms
            pos_sim: similarity scores between cand terms and pos opinion terms

        Returns:
            x: concatenated features and similarity scores
        """
        neg = np.array(neg_sim)
        pos = np.array(pos_sim)

        neg_avg = np.mean(neg, axis=1, keepdims=True)
        neg_std = np.std(neg, axis=1, keepdims=True)
        neg_min = np.min(neg, axis=1, keepdims=True)
        neg_max = np.max(neg, axis=1, keepdims=True)

        pos_avg = np.mean(pos, axis=1, keepdims=True)
        pos_std = np.std(pos, axis=1, keepdims=True)
        pos_min = np.min(pos, axis=1, keepdims=True)
        pos_max = np.max(pos, axis=1, keepdims=True)

        print('\nAdding polarity similarity features...')

        res_x = np.concatenate(
            (neg_avg, neg_std, neg_min, neg_max, pos_avg, pos_std, pos_min, pos_max, x), 1)

        return res_x

    def generate_embbeding_features(self, terms, embedding_dict):
        """
        concatenate word embedding to features

        Args:
            terms: candidate terms
            embedding_dict: embedding dictionary
            word_to_emb_idx: index to embedding dictionary
        Returns:
            x: concatenated features and word embs
        """
        print("\nAdding word vector features...\n")
        vec_matrix = np.zeros((len(terms), self.embeddings_len))

        j = 0
        for term in terms:
            word_vector = embedding_dict.get(term)
            vec_matrix[j, :] = word_vector
            j += 1

        x = vec_matrix[:j]

        return x

    def load_terms_and_y_labels_and_generate_features(self, filename):
        """
       load candidate terms with their basic features, Y labels and polarities from feature file

       Args:
           filename: feature table file path
       Returns:
           x: feature vector
           y: labels vector
           terms: candidate terms
           polarities: opinion polarity per term
       """
        print("\nLoading feature table...\n")

        y, terms, polarities = self.load_terms_and_y_labels(filename)

        x, terms, polarities = self.generate_features(terms, polarities)

        y_vector = None
        if y is not None:
            y_vector = np.reshape(y, (y.shape[0], 1))

        return x, y, y_vector, terms, polarities

    def load_terms_and_generate_features(self, filename):
        """
       load candidate terms with their basic features, Y labels and polarities from feature file

       Args:
           filename: feature table file path
       Returns:
           x: feature vector
           terms: candidate terms
           polarities: opinion polarity per term
       """
        print("\nLoading feature table...\n")

        terms, polarities = self.load_terms_and_polarities(filename)

        x, terms, polarities = self.generate_features(terms, polarities)

        return x, terms, polarities

    @staticmethod
    def _determine_unk_polarities(terms, polarities, neg, pos):

        for i, term in enumerate(terms):
            if np.average(pos[i]) <= np.average(neg[i]):
                polarities[term] = 'POS'
            else:
                polarities[term] = 'NEG'

        return polarities

    def generate_features(self, terms, polarities):

        generic_terms = _read_generic_lex_for_similarity(self.generic_lexicon)
        # generate unified list of candidate terms and generic terms
        terms_list = [term for term in terms]
        for term in generic_terms.keys():
            terms_list.append(term.strip('\'"'))

        print("\nLoading embedding model...\n")
        embedding_dict, _ = load_word_embeddings(self.emb_model_path, terms_list)
        x = self.generate_embbeding_features(terms, embedding_dict)
        neg, pos = self.calc_similarity_scores_for_all_terms(terms, generic_terms, embedding_dict)
        x = self.concat_sim_scores_and_features(x, neg, pos)
        polarities = self._determine_unk_polarities(terms, polarities, neg, pos)
        print("\nDimensions of X: " + str(x.shape))
        return x, terms, polarities

    def evaluate(self, model, x_test, y_test, terms):
        report = {}
        predictions = model.predict(x_test, verbose=0)
        tp = 0
        fp = 0
        tn = 0
        fn = 0

        for i, prediction in enumerate(predictions):
            y_true = y_test[i][0]

            if prediction[0] > self.threshold:
                y_pred = 1
            else:
                y_pred = 0

            report[terms[i]] = (prediction[0], y_pred, y_true)

            if y_pred == 1:
                if y_true == 1:
                    tp = tp + 1
                else:
                    fp = fp + 1
            elif y_true == 0:
                tn = tn + 1
            else:
                fn = fn + 1

        prec = 100 * tp / (tp + fp)
        rec = 100 * tp / (tp + fn)
        f1 = 2 * (prec * rec) / (prec + rec)

        return (prec, rec, f1), report

    def generate_model(self, input_vector_dimension):
        """Generate MLP model.

        Args:
           input_vector_dimension (int): word emb vec length

        Returns:
        """
        mlp_model = Sequential()

        mlp_model.add(Dense(128, activation=self.activation_1, input_dim=input_vector_dimension))
        mlp_model.add(Dropout(0.5))
        mlp_model.add(Dense(64, activation=self.activation_2))
        mlp_model.add(Dropout(0.5))
        mlp_model.add(Dense(1, activation=self.activation_3))
        mlp_model.compile(metrics=['accuracy'], loss=self.loss, optimizer=self.optimizer)

        return mlp_model

    def predict(self, input_table_file, generic_opinion_terms):
        """Predict classification class according to model.

        Args:
           input_table_file: feature(X) and labels(Y) table file
           generic_opinion_terms: generic opinion terms file name

        Returns:
            final_concat_opinion_lex: reranked_lex conctenated with generic lex
        """
        x, terms, polarities = self.load_terms_and_generate_features(input_table_file)

        model = load_model(self.rerank_model_path)
        reranked_lexicon = model.predict(x, verbose=0)

        reranked_lex = {}
        for i, prediction in enumerate(reranked_lexicon):
            if not np.isnan(prediction[0]) and prediction[0] > self.PREDICTION_THRESHOLD:
                reranked_lex[terms[i]] = (prediction[0], polarities[terms[i]])

        final_concat_opinion_lex = \
            self._generate_concat_reranked_lex(reranked_lex, generic_opinion_terms)
        self._write_prediction_results(final_concat_opinion_lex)
        return final_concat_opinion_lex

    def rerank_train(self):
        """Class for training a reranking model."""
        x, y, _, _, _ = \
            self.load_terms_and_y_labels_and_generate_features(self.train_rerank_data_path)

        try:
            print('\nModel training...')
            model = self.generate_model(x.shape[1])
            e = self.epochs_and_batch_size[0][0]
            b = self.epochs_and_batch_size[0][1]

            model.fit(x, y, epochs=e, batch_size=b, verbose=0)
            RerankTerms.model_dir.mkdir(parents=True, exist_ok=True)

            model.save(str(RerankTerms.model_dir) + '/rerank_model.h5')
            print('\nSaved model to: ' + str(RerankTerms.model_dir) + '/rerank_model.h5')

        except ZeroDivisionError:
            print("Division by zero, skipping test")

    def cross_validation_training(self, verbose=False):
        """Perform k fold cross validation and evaluate the results."""
        final_report = {}
        x, y, y_vector, terms, _ = \
            self.load_terms_and_y_labels_and_generate_features(self.train_rerank_data_path)

        for seed in self.seeds:
            np.random.seed(seed)

            for epochs, batch_size in self.epochs_and_batch_size:
                self.print_params(batch_size, epochs, seed)
                k_fold = StratifiedKFold(n_splits=5, shuffle=True, random_state=seed)
                f1_scores = []
                precision_scores = []
                recall_scores = []

                try:

                    for i, (train, test) in enumerate(k_fold.split(x, y)):
                        model = self.generate_model(x.shape[1])
                        model.fit(x[train], y_vector[train], epochs=epochs, batch_size=batch_size,
                                  verbose=0)

                        measures, report = self.evaluate(model, x[test], y_vector[test],
                                                         terms[test])
                        final_report.update(report)

                        precision, recall, f1 = measures
                        f1_scores.append(f1)
                        precision_scores.append(precision)
                        recall_scores.append(recall)

                        if verbose:
                            print("Fold " + str(i + 1) + ":")
                            self.print_evaluation_results(precision, recall, f1)

                    print('\nSummary:')
                    self.print_evaluation_results(precision_scores, recall_scores, f1_scores)

                except ZeroDivisionError:
                    print("Division by zero, skipping test")

        self.write_evaluation_report(final_report)

    def print_params(self, batch_size, epochs, seed):
        """Print training params.

        Args:
            batch_size(int): batch size
            epochs(int): num of epochs
            seed(int): seed
        """
        print('\nModel Parameters: act_1= ' + self.activation_1 + ', act_2= ' + self.activation_2
              + ', act_3= ' + self.activation_3 + ', loss= ' + self.loss + ', optimizer= '
              + self.optimizer + '\nseed= ' + str(seed) + ', epochs= ' + str(epochs)
              + ', batch_size= ' + str(batch_size) + ', threshold= ' + str(self.threshold)
              + ', use_complete_w2v= ' + ', sim_lexicon= ' + str(self.sim_lexicon) + '\n')

    def print_evaluation_results(self, precision, recall, f1):
        """Print evaluation results.

        Args:
            precision(list of float): precision
            recall(list of float): recall
            f1(list of float): f measure
        """
        print()
        self.print_measure('Precision', precision)
        self.print_measure('Recall', recall)
        self.print_measure('F-measure', f1)
        print('-------------------------------------------------------------------------'
              '------------------------------')

    @staticmethod
    def print_measure(measure, value):
        """Print single measure.

        Args:
            measure(str): measure type
            value(list of float): value
        """
        print(measure + ': {:.2f}%'.format(np.mean(value)), end='')
        if not np.isscalar(value):
            print(" (+/- {:.2f}%)".format(np.std(value)), end='')
        print()

    @staticmethod
    def _generate_concat_reranked_lex(acquired_opinion_lex, generic_opinion_lex_file):
        print('Loading generic sentiment terms from {}...'.format(generic_opinion_lex_file))
        generics_table = np.genfromtxt(generic_opinion_lex_file, delimiter=',', skip_header=1,
                                       dtype=str)
        print(str(generics_table.shape[0]) + ' generic sentiment terms loaded')

        concat_opinion_dict = {}

        for key, value in acquired_opinion_lex.items():
            concat_opinion_dict[key] = (value[0], value[1], 'Y')
        for row in generics_table:
            concat_opinion_dict[row[0]] = (row[2], row[1], 'N')

        return concat_opinion_dict

    @staticmethod
    def _write_prediction_results(concat_opinion_dict):

        out_path = RerankTerms.out_dir / 'generated_opinion_lex_reranked.csv'
        with open(out_path, 'w') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(['Term', 'Score', 'Polarity', 'isAcquired'])
            for key, value in concat_opinion_dict.items():
                writer.writerow([key, value[0], value[1], value[2]])
        print('Reranked opinion lexicon written to {}'.format(out_path))

    @staticmethod
    def write_evaluation_report(report_dic):
        RerankTerms.model_dir.mkdir(parents=True, exist_ok=True)
        out_path = RerankTerms.model_dir / 'rerank_classifier_results.csv'
        with open(out_path, 'w') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(['term', 'score', 'y_pred', 'y_true'])
            for key, value in report_dic.items():
                writer.writerow([key, value[0], value[1], value[2]])
        print('Report written to {}' + str(out_path))

    @staticmethod
    def load_word_vectors_dict():
        try:
            with open(RerankTerms.model_dir / 'word_vectors_dict.pickle', 'rb') as f:
                ret = pickle.load(f)
        except OSError:
            ret = {}
        return ret



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import json
import logging

from nlp_architect.common.cdc.mention_data import MentionData
from nlp_architect.data.cdc_resources.wordnet.wordnet_online import WordnetOnline
from nlp_architect.utils import io
from nlp_architect.utils.io import json_dumper

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(description='Create WordNet dataset only dump')

parser.add_argument('--mentions', type=str, help='mentions file', required=True)

parser.add_argument('--output', type=str, help='location were to create dump file', required=True)

args = parser.parse_args()


def wordnet_dump():
    out_file = args.output
    mentions_file = args.mentions
    logger.info('Loading mentions files...')
    mentions = MentionData.read_mentions_json_to_mentions_data_list(mentions_file)
    logger.info('Done loading mentions files, starting local dump creation...')
    result_dump = dict()
    wordnet = WordnetOnline()
    for mention in mentions:
        page = wordnet.get_pages(mention)
        result_dump[page.orig_phrase] = page

    with open(out_file, 'w') as out:
        json.dump(result_dump, out, default=json_dumper)

    logger.info('Wordnet Dump Created Successfully, '
                'extracted total of %d wn pages', len(result_dump))
    logger.info('Saving dump to file-%s', out_file)


if __name__ == '__main__':
    io.validate_existing_filepath(args.mentions)
    io.validate_existing_filepath(args.output)
    wordnet_dump()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

MAJOR_V = 0
MINOR_V = 4
PATCH_V = 0
STAGE = ''


def nlp_architect_version():
    if PATCH_V != 0:
        v = '{}.{}.{}'.format(MAJOR_V, MINOR_V, PATCH_V)
    else:
        v = '{}.{}'.format(MAJOR_V, MINOR_V)
    if len(STAGE) != 0:
        v += '-{}'.format(STAGE)
    return v


NLP_ARCHITECT_VERSION = nlp_architect_version()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ******************************************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""Attention-based sequence-to-sequence model with dynamic RNN support."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from nlp_architect.models.gnmt import model
from nlp_architect.models.gnmt import model_helper

__all__ = ["AttentionModel"]


class AttentionModel(model.Model):
    """Sequence-to-sequence dynamic model with attention.

    This class implements a multi-layer recurrent neural network as encoder,
    and an attention-based decoder. This is the same as the model described in
    (Luong et al., EMNLP'2015) paper: https://arxiv.org/pdf/1508.04025v5.pdf.
    This class also allows to use GRU cells in addition to LSTM cells with
    support for dropout.
    """

    def __init__(self,
                 hparams,
                 mode,
                 iterator,
                 source_vocab_table,
                 target_vocab_table,
                 reverse_target_vocab_table=None,
                 scope=None,
                 extra_args=None):
        self.has_attention = hparams.attention_architecture and hparams.attention

        # Set attention_mechanism_fn
        if self.has_attention:
            if extra_args and extra_args.attention_mechanism_fn:
                self.attention_mechanism_fn = extra_args.attention_mechanism_fn
            else:
                self.attention_mechanism_fn = create_attention_mechanism

        super(AttentionModel, self).__init__(
            hparams=hparams,
            mode=mode,
            iterator=iterator,
            source_vocab_table=source_vocab_table,
            target_vocab_table=target_vocab_table,
            reverse_target_vocab_table=reverse_target_vocab_table,
            scope=scope,
            extra_args=extra_args)

    def _prepare_beam_search_decoder_inputs(
            self, beam_width, memory, source_sequence_length, encoder_state):
        memory = tf.contrib.seq2seq.tile_batch(
            memory, multiplier=beam_width)
        source_sequence_length = tf.contrib.seq2seq.tile_batch(
            source_sequence_length, multiplier=beam_width)
        encoder_state = tf.contrib.seq2seq.tile_batch(
            encoder_state, multiplier=beam_width)
        batch_size = self.batch_size * beam_width
        return memory, source_sequence_length, encoder_state, batch_size

    def _build_decoder_cell(self, hparams, encoder_outputs, encoder_state,
                            source_sequence_length):
        """Build a RNN cell with attention mechanism that can be used by decoder."""
        # No Attention
        if not self.has_attention:
            return super(AttentionModel, self)._build_decoder_cell(
                hparams, encoder_outputs, encoder_state, source_sequence_length)
        elif hparams.attention_architecture != "standard":
            raise ValueError(
                "Unknown attention architecture %s" % hparams.attention_architecture)

        num_units = hparams.num_units
        num_layers = self.num_decoder_layers
        num_residual_layers = self.num_decoder_residual_layers
        infer_mode = hparams.infer_mode

        dtype = tf.float32

        # Ensure memory is batch-major
        if self.time_major:
            memory = tf.transpose(encoder_outputs, [1, 0, 2])
        else:
            memory = encoder_outputs

        if self.mode == tf.contrib.learn.ModeKeys.INFER and infer_mode == "beam_search":
            memory, source_sequence_length, encoder_state, batch_size = (
                self._prepare_beam_search_decoder_inputs(
                    hparams.beam_width, memory, source_sequence_length,
                    encoder_state))
        else:
            batch_size = self.batch_size

        # Attention
        attention_mechanism = self.attention_mechanism_fn(
            hparams.attention, num_units, memory, source_sequence_length, self.mode)

        cell = model_helper.create_rnn_cell(
            unit_type=hparams.unit_type,
            num_units=num_units,
            num_layers=num_layers,
            num_residual_layers=num_residual_layers,
            forget_bias=hparams.forget_bias,
            dropout=hparams.dropout,
            num_gpus=self.num_gpus,
            mode=self.mode,
            single_cell_fn=self.single_cell_fn)

        # Only generate alignment in greedy INFER mode.
        alignment_history = (self.mode == tf.contrib.learn.ModeKeys.INFER
                             and infer_mode != "beam_search")
        cell = tf.contrib.seq2seq.AttentionWrapper(
            cell,
            attention_mechanism,
            attention_layer_size=num_units,
            alignment_history=alignment_history,
            output_attention=hparams.output_attention,
            name="attention")

        # TODO(thangluong): do we need num_layers, num_gpus?
        cell = tf.contrib.rnn.DeviceWrapper(cell,
                                            model_helper.get_device_str(
                                                num_layers - 1, self.num_gpus))

        if hparams.pass_hidden_state:
            decoder_initial_state = cell.zero_state(batch_size, dtype).clone(
                cell_state=encoder_state)
        else:
            decoder_initial_state = cell.zero_state(batch_size, dtype)

        return cell, decoder_initial_state

    def _get_infer_summary(self, hparams):
        if not self.has_attention or hparams.infer_mode == "beam_search":
            return tf.no_op()
        return _create_attention_images_summary(self.final_context_state)


def create_attention_mechanism(attention_option, num_units, memory,
                               source_sequence_length, mode):
    """Create attention mechanism based on the attention_option."""
    del mode  # unused

    # Mechanism
    if attention_option == "luong":
        attention_mechanism = tf.contrib.seq2seq.LuongAttention(
            num_units, memory, memory_sequence_length=source_sequence_length)
    elif attention_option == "scaled_luong":
        attention_mechanism = tf.contrib.seq2seq.LuongAttention(
            num_units,
            memory,
            memory_sequence_length=source_sequence_length,
            scale=True)
    elif attention_option == "bahdanau":
        attention_mechanism = tf.contrib.seq2seq.BahdanauAttention(
            num_units, memory, memory_sequence_length=source_sequence_length)
    elif attention_option == "normed_bahdanau":
        attention_mechanism = tf.contrib.seq2seq.BahdanauAttention(
            num_units,
            memory,
            memory_sequence_length=source_sequence_length,
            normalize=True)
    else:
        raise ValueError("Unknown attention option %s" % attention_option)

    return attention_mechanism


def _create_attention_images_summary(final_context_state):
    """create attention image and attention summary."""
    attention_images = (final_context_state.alignment_history.stack())
    # Reshape to (batch, src_seq_len, tgt_seq_len,1)
    attention_images = tf.expand_dims(
        tf.transpose(attention_images, [1, 2, 0]), -1)
    # Scale to range [0, 255]
    attention_images *= 255
    attention_summary = tf.summary.image("attention_images", attention_images)
    return attention_summary



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import traceback

import requests
from elasticsearch import Elasticsearch

from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page import WikipediaPage
from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page_extracted_relations import \
    WikipediaPageExtractedRelations
from nlp_architect.data.cdc_resources.wikipedia.wiki_search_page_result import \
    WikipediaSearchPageResult


class WikiElastic(object):
    def __init__(self, host: str, port: int, index: str):
        # connect to our cluster
        self.cache = dict()
        if self.is_connected(host, port):
            self.es_index = index
            self.es = Elasticsearch([{'host': host, 'port': port}])
        else:
            traceback.print_exc()
            raise IOError('Cannot connect to ElasticSearch node')

    @staticmethod
    def is_connected(elastic_host, elastic_port):
        elastic_search_url = 'http://' + elastic_host + ':' + str(elastic_port)
        res = requests.get(elastic_search_url)
        if res.content:
            return True
        return False

    def get_pages(self, phrase):
        if phrase in self.cache:
            return self.cache[phrase]
        try:
            phrase_strip = ' '.join(phrase.replace('-', ' ').split())
            pages = set()
            best_results = self.get_best_elastic_results(phrase_strip)
            for result in best_results:
                _id = result['_id']
                if _id != 0:
                    result_source = result['_source']
                    if 'redirectTitle' in result_source:
                        redirect_title = result_source['redirectTitle']
                        red_result = None
                        while redirect_title and result_source['title'] != redirect_title:
                            red_result = self.get_redirect_result(redirect_title)
                            if red_result is None or len(red_result) == 0:
                                print('could not find redirect title=' + redirect_title
                                      + ', does not exist in data')
                                redirect_title = None
                            elif 'redirectTitle' in red_result[0]['_source']:
                                redirect_title = red_result[0]['_source']['redirectTitle']
                            else:
                                redirect_title = None

                        if red_result is not None and len(red_result) > 0:
                            result = red_result[0]
                            _id = result['_id']

                    elastic_page_result = self.get_page_from_result_v1(phrase_strip, result, _id)
                    pages.add(WikipediaSearchPageResult(phrase, elastic_page_result))

            self.cache[phrase] = pages
            return pages
        except Exception:
            traceback.print_exc()

    def get_best_elastic_results(self, phrase):
        best_results = []
        best_results.extend(self.get_redirect_result(phrase))

        search_result_near_match = self.es.search(index=self.es_index,
                                                  body={"size": 5,
                                                        'query': {'match_phrase': {
                                                            'title.near_match': phrase}}})
        best_results.extend(self.extract_from_elastic_results(search_result_near_match))

        return best_results

    @staticmethod
    def extract_from_elastic_results(search_result):
        best_results = []
        if search_result is not None and search_result['hits']['total'] > 0:
            if search_result['hits']['total'] > 0:
                best_results = search_result['hits']['hits']
        return best_results

    def get_redirect_result(self, phrase):
        search_result = self.es.search(index=self.es_index,
                                       body={"size": 5,
                                             'query': {'match_phrase': {'title.keyword': phrase}}})
        results = self.extract_from_elastic_results(search_result)
        return results

    def get_page_from_result_v1(self, phrase, result, result_id):
        if result_id != 0 and result is not None:
            relations = None
            result_source = result['_source']
            result_score = result['_score']
            if result_source is not None:
                title = result_source['title']
                relations_source = result_source['relations']

                if relations_source is not None:
                    is_part = relations_source['isPartName']
                    is_disambig = relations_source['isDisambiguation']

                    disambig_links = self.safe_extract_field_from_dict('disambiguationLinks',
                                                                       relations_source)
                    disambig_links_norm = self.safe_extract_field_from_dict(
                        'disambiguationLinksNorm', relations_source)
                    categories = self.safe_extract_field_from_dict('categories', relations_source)
                    categories_norm = self.safe_extract_field_from_dict('categoriesNorm',
                                                                        relations_source)
                    title_parent = self.safe_extract_field_from_dict('titleParenthesis',
                                                                     relations_source)
                    title_parent_norm = self.safe_extract_field_from_dict('titleParenthesisNorm',
                                                                          relations_source)
                    be_comp = self.safe_extract_field_from_dict('beCompRelations',
                                                                relations_source)
                    be_comp_norm = self.safe_extract_field_from_dict('beCompRelationsNorm',
                                                                     relations_source)

                    relations = WikipediaPageExtractedRelations(
                        is_part, is_disambig, title_parent, disambig_links,
                        categories, None, be_comp, disambig_links_norm, categories_norm, None,
                        title_parent_norm,
                        be_comp_norm
                    )

            return WikipediaPage(phrase, None, title, None, result_score, result_id, None,
                                 relations)

        return WikipediaPage()

    @staticmethod
    def safe_extract_field_from_dict(field_name, _dict):
        if field_name in _dict:
            return _dict[field_name]
        return None



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import tensorflow as tf


class CRF(tf.keras.layers.Layer):
    """
    Conditional Random Field layer (tf.keras)
    `CRF` can be used as the last layer in a network (as a classifier). Input shape (features)
    must be equal to the number of classes the CRF can predict (a linear layer is recommended).

    Note: the loss and accuracy functions of networks using `CRF` must
    use the provided loss and accuracy functions (denoted as loss and viterbi_accuracy)
    as the classification of sequences are used with the layers internal weights.

    Args:
        num_labels (int): the number of labels to tag each temporal input.

    Input shape:
        nD tensor with shape `(batch_size, sentence length, num_classes)`.

    Output shape:
        nD tensor with shape: `(batch_size, sentence length, num_classes)`.
    """

    def __init__(self, num_classes, **kwargs):
        self.transitions = None
        super(CRF, self).__init__(**kwargs)
        # num of output labels
        self.output_dim = int(num_classes)
        self.input_spec = tf.keras.layers.InputSpec(min_ndim=3)
        self.supports_masking = False
        self.sequence_lengths = None

    def get_config(self):
        config = {
            'output_dim': self.output_dim,
            'supports_masking': self.supports_masking,
            'transitions': tf.keras.backend.eval(self.transitions)
        }
        base_config = super(CRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def build(self, input_shape):
        assert len(input_shape) == 3
        f_shape = tf.TensorShape(input_shape)
        input_spec = tf.keras.layers.InputSpec(min_ndim=3, axes={-1: f_shape[-1]})

        if f_shape[-1] is None:
            raise ValueError('The last dimension of the inputs to `CRF` '
                             'should be defined. Found `None`.')
        if f_shape[-1] != self.output_dim:
            raise ValueError('The last dimension of the input shape must be equal to output'
                             ' shape. Use a linear layer if needed.')
        self.input_spec = input_spec
        self.transitions = self.add_weight(name='transitions',
                                           shape=[self.output_dim, self.output_dim],
                                           initializer='glorot_uniform',
                                           trainable=True)
        self.built = True

    # pylint: disable=arguments-differ
    def call(self, inputs, sequence_lengths=None, **kwargs):
        sequences = tf.convert_to_tensor(inputs, dtype=self.dtype)
        if sequence_lengths is not None:
            assert len(sequence_lengths.shape) == 2
            assert tf.convert_to_tensor(sequence_lengths).dtype == 'int32'
            seq_len_shape = tf.convert_to_tensor(sequence_lengths).get_shape().as_list()
            assert seq_len_shape[1] == 1
            self.sequence_lengths = tf.keras.backend.flatten(sequence_lengths)
        else:
            self.sequence_lengths = tf.ones(tf.shape(inputs)[0], dtype=tf.int32) * \
                (tf.shape(inputs)[1])

        viterbi_sequence, _ = tf.contrib.crf.crf_decode(sequences, self.transitions,
                                                        self.sequence_lengths)
        output = tf.keras.backend.one_hot(viterbi_sequence, self.output_dim)
        return tf.keras.backend.in_train_phase(sequences, output)

    def loss(self, y_true, y_pred):
        y_pred = tf.convert_to_tensor(y_pred, dtype=self.dtype)
        log_likelihood, self.transitions = \
            tf.contrib.crf.crf_log_likelihood(y_pred,
                                              tf.cast(tf.keras.backend.argmax(y_true),
                                                      dtype=tf.int32),
                                              self.sequence_lengths,
                                              transition_params=self.transitions)
        return tf.reduce_mean(-log_likelihood)

    def compute_output_shape(self, input_shape):
        tf.TensorShape(input_shape).assert_has_rank(3)
        return input_shape[:2] + (self.output_dim,)

    @property
    def viterbi_accuracy(self):
        def accuracy(y_true, y_pred):
            shape = tf.shape(y_pred)
            sequence_lengths = tf.ones(shape[0], dtype=tf.int32) * (shape[1])
            viterbi_sequence, _ = tf.contrib.crf.crf_decode(y_pred, self.transitions,
                                                            sequence_lengths)
            output = tf.keras.backend.one_hot(viterbi_sequence, self.output_dim)
            return tf.keras.metrics.categorical_accuracy(y_true, output)
        accuracy.func_name = 'viterbi_accuracy'
        return accuracy
